# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
'''
Resnet
'''
import numpy as np
import mindspore.nn as nn
import mindspore as ms
import mindspore.ops as ops

from MindsporeTrainer.modeling.loss import SoftmaxCrossEntropyExpand
from MindsporeTrainer.modeling.modeling_adapter import NetworkWithLoss


def weight_variable_0(shape):
    """weight_variable_0"""
    zeros = np.zeros(shape).astype(np.float32)
    return ms.Tensor(zeros)


def weight_variable_1(shape):
    """weight_variable_1"""
    ones = np.ones(shape).astype(np.float32)
    return ms.Tensor(ones)


def conv3x3(in_channels, out_channels, stride=1, padding=0):
    """3x3 convolution """
    return nn.Conv2d(in_channels, out_channels,
                     kernel_size=3, stride=stride, padding=padding, weight_init='XavierUniform',
                     has_bias=False, pad_mode="same")


def conv1x1(in_channels, out_channels, stride=1, padding=0):
    """1x1 convolution"""
    return nn.Conv2d(in_channels, out_channels,
                     kernel_size=1, stride=stride, padding=padding, weight_init='XavierUniform',
                     has_bias=False, pad_mode="same")


def conv7x7(in_channels, out_channels, stride=1, padding=0):
    """1x1 convolution"""
    return nn.Conv2d(in_channels, out_channels,
                     kernel_size=7, stride=stride, padding=padding, weight_init='XavierUniform',
                     has_bias=False, pad_mode="same")


def bn_with_initialize(out_channels):
    """bn_with_initialize"""
    shape = (out_channels)
    mean = weight_variable_0(shape)
    var = weight_variable_1(shape)
    beta = weight_variable_0(shape)
    bn = nn.BatchNorm2d(out_channels, momentum=0.99, eps=0.00001, gamma_init='Uniform',
                        beta_init=beta, moving_mean_init=mean, moving_var_init=var)
    return bn


def bn_with_initialize_last(out_channels):
    """bn_with_initialize_last"""
    shape = (out_channels)
    mean = weight_variable_0(shape)
    var = weight_variable_1(shape)
    beta = weight_variable_0(shape)
    bn = nn.BatchNorm2d(out_channels, momentum=0.99, eps=0.00001, gamma_init='Uniform',
                        beta_init=beta, moving_mean_init=mean, moving_var_init=var)
    return bn


def fc_with_initialize(input_channels, out_channels):
    """fc_with_initialize"""
    return nn.Dense(input_channels, out_channels, weight_init='XavierUniform', bias_init='Uniform')


class ResidualBlock(nn.Cell):
    """ResidualBlock"""
    expansion = 4

    def __init__(self,
                 in_channels,
                 out_channels,
                 stride=1):
        """init block"""
        super(ResidualBlock, self).__init__()

        out_chls = out_channels // self.expansion
        self.conv1 = conv1x1(in_channels, out_chls, stride=stride, padding=0)
        self.bn1 = bn_with_initialize(out_chls)

        self.conv2 = conv3x3(out_chls, out_chls, stride=1, padding=0)
        self.bn2 = bn_with_initialize(out_chls)

        self.conv3 = conv1x1(out_chls, out_channels, stride=1, padding=0)
        self.bn3 = bn_with_initialize_last(out_channels)

        self.relu = ops.ReLU()
        self.add = ops.Add()

    def construct(self, x):
        """construct"""
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out = self.add(out, identity)
        out = self.relu(out)

        return out


class ResidualBlockWithDown(nn.Cell):
    """ResidualBlockWithDown"""
    expansion = 4

    def __init__(self,
                 in_channels,
                 out_channels,
                 stride=1,
                 down_sample=False):
        """init block with down"""
        super(ResidualBlockWithDown, self).__init__()

        out_chls = out_channels // self.expansion
        self.conv1 = conv1x1(in_channels, out_chls, stride=stride, padding=0)
        self.bn1 = bn_with_initialize(out_chls)

        self.conv2 = conv3x3(out_chls, out_chls, stride=1, padding=0)
        self.bn2 = bn_with_initialize(out_chls)

        self.conv3 = conv1x1(out_chls, out_channels, stride=1, padding=0)
        self.bn3 = bn_with_initialize_last(out_channels)

        self.relu = ops.ReLU()
        self.down_sample = down_sample

        self.conv_down_sample = conv1x1(in_channels, out_channels, stride=stride, padding=0)
        self.bn_down_sample = bn_with_initialize(out_channels)
        self.add = ops.Add()

    def construct(self, x):
        """construct"""
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        identity = self.conv_down_sample(identity)
        identity = self.bn_down_sample(identity)

        out = self.add(out, identity)
        out = self.relu(out)

        return out


class MakeLayer0(nn.Cell):
    """MakeLayer0"""

    def __init__(self, block, in_channels, out_channels, stride):
        """init"""
        super(MakeLayer0, self).__init__()
        self.a = ResidualBlockWithDown(in_channels, out_channels, stride=1, down_sample=True)
        self.b = block(out_channels, out_channels, stride=stride)
        self.c = block(out_channels, out_channels, stride=1)

    def construct(self, x):
        """construct"""
        x = self.a(x)
        x = self.b(x)
        x = self.c(x)

        return x


class MakeLayer1(nn.Cell):
    """MakeLayer1"""

    def __init__(self, block, in_channels, out_channels, stride):
        """init"""
        super(MakeLayer1, self).__init__()
        self.a = ResidualBlockWithDown(in_channels, out_channels, stride=stride, down_sample=True)
        self.b = block(out_channels, out_channels, stride=1)
        self.c = block(out_channels, out_channels, stride=1)
        self.d = block(out_channels, out_channels, stride=1)

    def construct(self, x):
        """construct"""
        x = self.a(x)
        x = self.b(x)
        x = self.c(x)
        x = self.d(x)

        return x


class MakeLayer2(nn.Cell):
    """MakeLayer2"""

    def __init__(self, block, in_channels, out_channels, stride):
        """init"""
        super(MakeLayer2, self).__init__()
        self.a = ResidualBlockWithDown(in_channels, out_channels, stride=stride, down_sample=True)
        self.b = block(out_channels, out_channels, stride=1)
        self.c = block(out_channels, out_channels, stride=1)
        self.d = block(out_channels, out_channels, stride=1)
        self.e = block(out_channels, out_channels, stride=1)
        self.f = block(out_channels, out_channels, stride=1)

    def construct(self, x):
        """construct"""
        x = self.a(x)
        x = self.b(x)
        x = self.c(x)
        x = self.d(x)
        x = self.e(x)
        x = self.f(x)

        return x


class MakeLayer3(nn.Cell):
    """MakeLayer3"""

    def __init__(self, block, in_channels, out_channels, stride):
        """init"""
        super(MakeLayer3, self).__init__()
        self.a = ResidualBlockWithDown(in_channels, out_channels, stride=stride, down_sample=True)
        self.b = block(out_channels, out_channels, stride=1)
        self.c = block(out_channels, out_channels, stride=1)

    def construct(self, x):
        """construct"""
        x = self.a(x)
        x = self.b(x)
        x = self.c(x)

        return x


class Head(nn.Cell):
    """Head"""
    def __init__(self):
        super(Head, self).__init__()
        self.conv1 = conv7x7(3, 64, stride=2, padding=0)
        self.bn1 = bn_with_initialize(64)
        self.relu = ops.ReLU()
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, pad_mode="same")

    def construct(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        return x


class ResNet(nn.Cell):
    """ResNet"""

    def __init__(self, block, num_classes=100, batch_size=32):
        """init"""
        super(ResNet, self).__init__()
        self.batch_size = batch_size
        self.num_classes = num_classes
        self.head = Head()

        self.layer1 = MakeLayer0(block, in_channels=64, out_channels=256, stride=1)
        self.layer2 = MakeLayer1(block, in_channels=256, out_channels=512, stride=2)
        self.layer3 = MakeLayer2(block, in_channels=512, out_channels=1024, stride=2)
        self.layer4 = MakeLayer3(block, in_channels=1024, out_channels=2048, stride=2)

        self.pool = ops.ReduceMean(keep_dims=True)
        self.squeeze = ops.Squeeze(axis=(2, 3))
        self.fc = fc_with_initialize(512 * block.expansion, num_classes)

        # pipeline parallel config
        self.head.pipeline_stage = 0
        self.layer1.pipeline_stage = 0
        self.layer2.pipeline_stage = 0
        self.layer3.pipeline_stage = 1
        self.layer4.pipeline_stage = 1
        self.fc.pipeline_stage = 1

    def construct(self, *sample):
        """construct"""
        x = sample[0]
        x = self.head(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.pool(x, (2, 3))
        x = self.squeeze(x)
        x = self.fc(x)
        return (x,)


def resnet50(batch_size, num_classes, eval_model=False):
    """create resnet50"""
    net = ResNet(ResidualBlock, num_classes, batch_size)
    return NetworkWithLoss(net, nn.SoftmaxCrossEntropyWithLogits(sparse=True, reduction='mean'), eval_model) 
    

class SoftmaxCrossEntropyExpand(nn.Cell):       # pylint: disable=missing-docstring
    def __init__(self, sparse=False):
        super(SoftmaxCrossEntropyExpand, self).__init__()
        self.exp = ops.Exp()
        self.sum = ops.ReduceSum(keep_dims=True)
        self.onehot = ops.OneHot()
        self.on_value = ms.Tensor(1.0, ms.float32)
        self.off_value = ms.Tensor(0.0, ms.float32)
        self.div = ops.RealDiv()
        self.log = ops.Log()
        self.sum_cross_entropy = ops.ReduceSum(keep_dims=False)
        self.mul = ops.Mul()
        self.mul2 = ops.Mul()
        self.mean = ops.ReduceMean(keep_dims=False)
        self.sparse = sparse
        self.max = ops.ReduceMax(keep_dims=True)
        self.sub = ops.Sub()
        self.eps = ms.Tensor(1e-8, ms.float32)

    def construct(self, logit, label):      # pylint: disable=missing-docstring
        logit_max = self.max(logit, -1)
        exp = self.exp(self.sub(logit, logit_max))
        exp_sum = self.sum(exp, -1)
        softmax_result = self.div(exp, exp_sum)
        if self.sparse:
            label = self.onehot(label, ops.shape(logit)[1], self.on_value, self.off_value)

        softmax_result_log = self.log(softmax_result + self.eps)
        loss = self.sum_cross_entropy((self.mul(softmax_result_log, label)), -1)
        loss = self.mul2(ops.scalar_to_array(-1.0), loss)
        loss = self.mean(loss, -1)

        return loss


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from .task import Task, TransformerTask, EvalData
from .task_registry import register_task, load_tasks, get_task



# Copyright (c) Facebook, Inc. and its affiliates.

#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# WYWEB Authors
# Date: 01/15/2020
#

# This piece of code is derived from https://github.com/pytorch/fairseq/blob/master/fairseq/data/encoders/gpt2_bpe.py

# import torch
import unicodedata
import os
from .gpt2_bpe_utils import get_encoder,_is_control,_is_whitespace,_is_punctuation
from .cache_utils import load_vocab

__all__ = ['GPT2Tokenizer']

class GPT2Tokenizer(object):
  """ A wrapper of GPT2 tokenizer with similar interface as BERT tokenizer

  Args:
    
    vocab_file (:obj:`str`, optional):
      The local path of vocabulary package or the release name of vocabulary in `DeBERTa GitHub releases <https://github.com/microsoft/DeBERTa/releases>`_, \
          e.g. "bpe_encoder", default: `None`. 
          
          If it's `None`, then it will download the vocabulary in the latest release from GitHub. The vocabulary file is a \
          state dictionary with three items, "dict_map", "vocab", "encoder" which correspond to three files used in `RoBERTa`, i.e. `dict.txt`, `vocab.txt` and `encoder.json`. \
          
          The difference between our wrapped GPT2 tokenizer and RoBERTa wrapped tokenizer are,

          - Special tokens, unlike `RoBERTa` which use `<s>`, `</s>` as the `start` token and `end` token of a sentence. We use `[CLS]` and `[SEP]` as the `start` and `end`\
              token of input sentence which is the same as `BERT`.

          - We remapped the token ids in our dictionary with regarding to the new special tokens, `[PAD]` => 0, `[CLS]` => 1, `[SEP]` => 2, `[UNK]` => 3, `[MASK]` => 50264

    do_lower_case (:obj:`bool`, optional):
      Whether to convert inputs to lower case. **Not used in GPT2 tokenizer**.

    special_tokens (:obj:`list`, optional):
      List of special tokens to be added to the end of the vocabulary.


  """
  def __init__(self, vocab_file=None, do_lower_case=True, special_tokens=None):
    self.pad_token='[PAD]'
    self.sep_token='[SEP]'
    self.unk_token='[UNK]'
    self.cls_token='[CLS]'

    self.symbols = []
    self.count = []
    self.indices = {}
    self.pad_token_id = self.add_symbol(self.pad_token)
    self.cls_token_id = self.add_symbol(self.cls_token)
    self.sep_token_id = self.add_symbol(self.sep_token)
    self.unk_token_id = self.add_symbol(self.unk_token)

    self.gpt2_encoder = torch.load(vocab_file)
    self.bpe = get_encoder(self.gpt2_encoder['encoder'], self.gpt2_encoder['vocab'])
    for w,n in self.gpt2_encoder['dict_map']:
      self.add_symbol(w, n)

    self.mask_token='[MASK]'
    self.mask_id = self.add_symbol(self.mask_token)
    self.special_tokens = ['[MASK]', '[SEP]', '[PAD]', '[UNK]', '[CLS]']
    if special_tokens is not None:
      for t in special_tokens:
        self.add_special_token(t)

    self.vocab = self.indices
    self.ids_to_tokens = self.symbols

  def tokenize(self, text):
    """ Convert an input text to tokens.
      
      Args:
        
        text (:obj:`str`): input text to be tokenized.

      Returns:
        A list of byte tokens where each token represent the byte id in GPT2 byte dictionary

      Example::
        
        >>> tokenizer = GPT2Tokenizer()
        >>> text = "Hello world!"
        >>> tokens = tokenizer.tokenize(text)
        >>> print(tokens)
        ['15496', '995', '0']
        
    """
    bpe = self._encode(text)

    return [t for t in bpe.split(' ') if t]

  def convert_tokens_to_ids(self, tokens):
    """ Convert list of tokens to ids.
      
      Args:

        tokens (:obj:`list<str>`): list of tokens

      Returns:
        
        List of ids
    """

    return [self.vocab[t] for t in tokens]

  def convert_ids_to_tokens(self, ids):
    """ Convert list of ids to tokens.
      
      Args:

        ids (:obj:`list<int>`): list of ids

      Returns:
        
        List of tokens
    """

    tokens = []
    for i in ids:
      tokens.append(self.ids_to_tokens[i])
    return tokens

  def split_to_words(self, text):
    return self.bpe.split_to_words(text)

  def decode(self, tokens):
    """ Decode list of tokens to text strings.
    
      Args:
        
        tokens (:obj:`list<str>`): list of tokens.

      Returns:
        
        Text string corresponds to the input tokens.

      Example::
        
        >>> tokenizer = GPT2Tokenizer()
        >>> text = "Hello world!"
        >>> tokens = tokenizer.tokenize(text)
        >>> print(tokens)
        ['15496', '995', '0']
        
        >>> tokenizer.decode(tokens)
        'Hello world!'
      
    """
    return self.bpe.decode([int(t) for t in tokens if t not in self.special_tokens])

  def add_special_token(self, token):
    """Adds a special token to the dictionary.
    
      Args:
        token (:obj:`str`): Tthe new token/word to be added to the vocabulary.

      Returns:
        The id of new token in the vocabulary.

    """
    self.special_tokens.append(token)
    return self.add_symbol(token)

  def part_of_whole_word(self, token, is_bos=False):
    if is_bos:
      return True
    s = self._decode(token)
    if (len(s)==1 and (_is_whitespace(list(s)[0]) or _is_control(list(s)[0]) or _is_punctuation(list(s)[0]))):
      return False

    return not s.startswith(' ')

  def sym(self, id):
    return self.ids_to_tokens[id]

  def id(self, sym):
    return self.vocab[sym]

  def _encode(self, x: str) -> str:
    return ' '.join(map(str, self.bpe.encode(x)))

  def _decode(self, x: str) -> str:
    return self.bpe.decode(map(int, x.split()))

  def add_symbol(self, word, n=1):
    """Adds a word to the dictionary.
    
      Args:
        word (:obj:`str`): Tthe new token/word to be added to the vocabulary.
        n (int, optional): The frequency of the word.

      Returns:
        The id of the new word.

    """
    if word in self.indices:
      idx = self.indices[word]
      self.count[idx] = self.count[idx] + n
      return idx
    else:
      idx = len(self.symbols)
      self.indices[word] = idx
      self.symbols.append(word)
      self.count.append(n)
      return idx

  def save_pretrained(self, path: str):
    torch.save(self.gpt2_encoder, path)



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
# This statement must be executed at the very beginning, i.e. before import torch
os.environ["OMP_NUM_THREADS"] = "1"



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import OrderedDict
import numpy as np
import numpy as np
import pickle

__all__=['ExampleInstance', 'example_to_feature', 'ExampleSet']

class ExampleInstance:
  def __init__(self, segments, label=None,  **kwv):
    self.segments = segments
    self.label = label
    self.__dict__.update(kwv)

  def __repr__(self):
    return f'segments: {self.segments}\nlabel: {self.label}'

  def __getitem__(self, i):
    return self.segments[i]

  def __len__(self):
    return len(self.segments)

class ExampleSet:
  def __init__(self, pairs):
    self._data = np.array([pickle.dumps(p) for p in pairs])
    self.total = len(self._data)

  def __getitem__(self, idx):
    """
    return pair
    """
    if isinstance(idx, tuple):
      idx,rng, ext_params = idx
    else:
      rng,ext_params=None, None
    content = self._data[idx]
    example = pickle.loads(content)
    segments = tuple([np.array(s) for s in example.segments])
    label = example.label
    if label is None:
      label = -1000
    return segments + (np.array([label]), )

  def __len__(self):
    return self.total

  def __iter__(self):
    for i in range(self.total):
      yield self[i]

def _truncate_segments(segments, max_num_tokens, rng):
  """
  Truncate sequence pair according to original BERT implementation:
  https://github.com/google-research/bert/blob/master/create_pretraining_data.py#L391
  """
  while True:
  #   if sum(len(s) for s in segments)<=max_num_tokens:
  #     break

    # segments = sorted(segments, key=lambda s:len(s), reverse=True)
    # trunc_tokens = segments[0]

    # assert len(trunc_tokens) >= 1
    if len(segments) <= max_num_tokens:
      break
    if rng.random() < 0.5:
      segments.pop(0)
    else:
      segments.pop()
  return segments

def truncate_tokens_pair(tokens_a, tokens_b, max_len, max_len_a=0, max_len_b=0, trunc_seg=None, always_truncate_tail=False):
    import random
    if isinstance(tokens_a, str) or isinstance(tokens_b, str):
      print('error')
    num_truncated_a = [0, 0]
    num_truncated_b = [0, 0]
    while True:
        if len(tokens_a) + len(tokens_b) <= max_len:
            break
        if (max_len_a > 0) and len(tokens_a) > max_len_a:
            trunc_tokens = tokens_a
            num_truncated = num_truncated_a
        elif (max_len_b > 0) and len(tokens_b) > max_len_b:
            trunc_tokens = tokens_b
            num_truncated = num_truncated_b
        elif trunc_seg:
            # truncate the specified segment
            if trunc_seg == 'a':
                trunc_tokens = tokens_a
                num_truncated = num_truncated_a
            else:
                trunc_tokens = tokens_b
                num_truncated = num_truncated_b
        else:
            # truncate the longer segment
            if len(tokens_a) > len(tokens_b):
                trunc_tokens = tokens_a
                num_truncated = num_truncated_a
            else:
                trunc_tokens = tokens_b
                num_truncated = num_truncated_b
        # whether always truncate source sequences
        if (not always_truncate_tail) and (random.random() < 0.5):
            del trunc_tokens[0]
            num_truncated[0] += 1
        else:
            trunc_tokens.pop()
            num_truncated[1] += 1
    return num_truncated_a, num_truncated_b




# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import numpy as np
import math
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score
from scipy.stats import pearsonr, spearmanr
from statistics import *
from scipy.special import softmax
from mindspore.nn.metrics import Metric, rearrange_inputs
# from mindspore.nn.metrics.metric import _check_onehot_data
from mindspore.nn.metrics.accuracy import Accuracy
import mindspore.ops.functional as F
import mindspore.common.dtype as mstype
from mindspore._checkparam import Validator as validator

def metric_multi_accuracy(logits, labels, options_num):
  logits = np.reshape(softmax(logits, -1)[:,1], (len(logits)//options_num, options_num))
  labels = np.argmax(np.reshape(labels, (len(labels)//options_num, options_num)),-1)
  return metric_accuracy(logits, labels)

def metric_accuracy(logits=None, labels=None, predicts=None):
  if logits is not None:
    predicts = np.argmax(logits, axis=1)
  return accuracy_score(labels, predicts)

def metric_precision(logits=None, labels=None, predicts=None, label_cls=None):
  if logits is not None:
    predicts = np.argmax(logits, axis=1)
  if labels.max() > 1:
    return precision_score(labels, predicts, average='macro', labels=label_cls)
  else:
    return precision_score(labels, predicts)

def metric_recall(logits=None, labels=None, predicts=None, label_cls=None):
  if logits is not None:
    predicts = np.argmax(logits, axis=1)
  if labels.max() > 1:
    return recall_score(labels, predicts, average='macro', labels=label_cls)
  else:
    return recall_score(labels, predicts)

def metric_f1(logits=None, labels=None, predicts=None, label_cls=None):
  if logits is not None:
    predicts = np.argmax(logits, axis=1)
  if labels.max() > 1:
    return f1_score(labels, predicts, average='macro', labels=label_cls)
  else:
    return f1_score(labels, predicts)

def metric_macro_f1(logits, ground_truth, labels=[0,1]):
  if logits is not None:
    predicts = np.argmax(logits, axis=1)
  f1=[]
  for l in labels:
    binary_g = (ground_truth==l).astype(np.int)
    binary_p = (predicts==l).astype(np.int)
    f1.append(f1_score(binary_g, binary_p))
  return float(np.mean(f1))

def metric_mcc(logits=None, labels=None, predicts=None):
  if logits is not None:
    predicts = np.argmax(logits, axis=1)
  return matthews_corrcoef(labels, predicts)


class BertMetric(Metric):
    """
    The metric of bert network.
    Args:
        batch_size (int): The batchsize of each device.
    """
    def __init__(self, batch_size):
        super(BertMetric, self).__init__()
        self.batch_size = batch_size
        self.ppl = Perplexity()
        self.clear()

    def clear(self):
        self.mlm_total = 0
        self.mlm_acc = 0
        self.loss = 0
        self.steps = 0
        self.ppl.clear()

    def update(self, *inputs):
        """Defines the computation performed.
            inputs: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                     "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"] + 
                    [prediction_scores, seq_relationship_score]
        """
        total_loss, acc, total, prediction_scores, masked_lm_ids = inputs
        mlm_acc = self._convert_data(acc)
        mlm_total = self._convert_data(total)
        self.mlm_acc += mlm_acc.sum()
        self.mlm_total += mlm_total.sum()
        self.loss += self._convert_data(total_loss.mean()).squeeze().tolist()
        self.steps += 1
        self.ppl.update(prediction_scores, masked_lm_ids)

    def eval(self):
        return {'acc':self.mlm_acc / self.mlm_total, 
                'loss': self.loss / self.steps, 
                'perplexity': self.ppl.eval()}


class MSAucuracy(Accuracy):
  def __call__(self, *inputs):
    """
    Evaluate input data once.

    Args:
        inputs (tuple): The first item is a predict array, the second item is a target array.

    Returns:
        Float, compute result.
    """
    if len(inputs) == 3:
      new_inputs = inputs[1:]
    self.clear()
    self.update(*new_inputs)
    return self.eval()

  def update(self, *inputs):
      if len(inputs) == 3:
        inputs = inputs[1:]
      if len(inputs) != 2:
          raise ValueError("For 'Accuracy.update', it needs 2 inputs (predicted value, true value), "
                            "but got {}".format(len(inputs)))
      y_pred = self._convert_data(inputs[0])
      y = self._convert_data(inputs[1])
      if self._type == 'classification' and y_pred.ndim != y.ndim:
          y_pred = y_pred.argmax(axis=1)
      # self._check_shape(y_pred, y)
      # self._check_value(y_pred, y)

      # if self._class_num == 0:
      #     self._class_num = y_pred.shape[1]
      # elif y_pred.shape[1] != self._class_num:
      #     raise ValueError("For 'Accuracy.update', class number not match, last input predicted data contain {} "
      #                       "classes, but current predicted data contain {} classes, please check your predicted "
      #                       "value(inputs[0]).".format(self._class_num, y_pred.shape[1]))

      if self._type == 'classification':
        if y_pred.ndim > 1:
          indices = y_pred.argmax(axis=1)
          result = (np.equal(indices, y) * 1).reshape(-1)
        else:
          result = (np.equal(y_pred, y) * 1).reshape(-1)
      elif self._type == 'multilabel':
          dimension_index = y_pred.ndim - 1
          y_pred = y_pred.swapaxes(1, dimension_index).reshape(-1, self._class_num)
          y = y.swapaxes(1, dimension_index).reshape(-1, self._class_num)
          result = np.equal(y_pred, y).all(axis=1) * 1

      self._correct_num += result.sum()
      self._total_num += result.shape[0]



class Perplexity(Metric):
    r"""
    Computes perplexity. Perplexity is a measurement about how well a probability distribution or a model predicts a
    sample. A low perplexity indicates the model can predict the sample well. The function is shown as follows:

    .. math::
        PP(W)=P(w_{1}w_{2}...w_{N})^{-\frac{1}{N}}=\sqrt[N]{\frac{1}{P(w_{1}w_{2}...w_{N})}}

    Args:
        ignore_label (int): Index of an invalid label to be ignored when counting. If set to `None`, it will include all
                            entries. Default: -1.

    Supported Platforms:
        ``Ascend`` ``GPU`` ``CPU``

    Note:
        The method `update` must be called with the form `update(preds, labels)`.

    Examples:
        >>> import numpy as np
        >>> from mindspore import nn, Tensor
        >>>
        >>> x = Tensor(np.array([[0.2, 0.5], [0.3, 0.1], [0.9, 0.6]]))
        >>> y = Tensor(np.array([1, 0, 1]))
        >>> metric = nn.Perplexity(ignore_label=None)
        >>> metric.clear()
        >>> metric.update(x, y)
        >>> perplexity = metric.eval()
        >>> print(perplexity)
        2.231443166940565
    """

    def __init__(self, ignore_label=-1):
        super(Perplexity, self).__init__()

        if ignore_label is None:
            self.ignore_label = ignore_label
        else:
            self.ignore_label = validator.check_value_type("ignore_label", ignore_label, [int])
        self.clear()

    def clear(self):
        """Clears the internal evaluation result."""
        self._sum_metric = 0.0
        self._num_inst = 0

    @rearrange_inputs
    def update(self, *inputs):
        """
        Updates the internal evaluation result: math:preds and :math:labels.

        Args:
            inputs: Input `preds` and `labels`. `preds` and `labels` are Tensor, list or numpy.ndarray.
                    `preds` is the predicted values, `labels` is the label of the data.
                    The shape of `preds` and `labels` are both :math:`(N, C)`.

        Raises:
            ValueError: If the number of the inputs is not 2.
            RuntimeError: If preds and labels have different lengths.
            RuntimeError: If label shape is not equal to pred shape.
        """
        if len(inputs) != 2:
            raise ValueError('The perplexity needs 2 inputs (preds, labels), but got {}.'.format(len(inputs)))

        preds = [self._convert_data(inputs[0])]
        labels = [self._convert_data(inputs[1])]

        if len(preds) != len(labels):
            raise RuntimeError('The preds and labels should have the same length, but the length of preds is{}, '
                               'the length of labels is {}.'.format(len(preds), len(labels)))

        loss = 0.
        num = 0
        for label, pred in zip(labels, preds):
          # pred scores are log_softmax, so it should be apply np.exp() to get probability
          pred = np.exp(pred)
          if label.size != pred.size / pred.shape[-1]:
              raise RuntimeError("shape mismatch: label shape should be equal to pred shape, but got label shape "
                                  "is {}, pred shape is {}.".format(label.shape, pred.shape))
          label = label.reshape((label.size,))
          label_expand = label.astype(int)
          label_expand = np.expand_dims(label_expand, axis=1)
          first_indices = np.arange(label_expand.shape[0])[:, None]
          pred = pred.reshape(-1, pred.shape[-1])
          pred = np.squeeze(pred[first_indices, label_expand]).astype(np.float32)
          if self.ignore_label is not None:
              ignore = (label == self.ignore_label).astype(pred.dtype)
              num -= np.sum(ignore)
              pred = pred * (1 - ignore) + ignore
          loss -= np.sum(np.log(np.maximum(1e-10, pred)))
          num += pred.size
        self._sum_metric += loss
        self._num_inst += num

    def eval(self):
        r"""
        Returns the current evaluation result.

        Returns:
            float, the computed result.

        Raises:
            RuntimeError: If the sample size is 0.
        """
        if self._num_inst == 0:
            raise RuntimeError('The perplexity can not be calculated, because the number of samples is 0.')

        return math.exp(self._sum_metric / self._num_inst)



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from mindspore.ops import operations as P
from mindspore.common import dtype as mstype

expand_dims = P.ExpandDims()
gather = P.Gather()
onehot = P.OneHot()
matmul = P.MatMul()
reshape = P.Reshape()
strided_slice = P.StridedSlice()
add = P.Add()
cast = P.Cast()
tile = P.Tile()
bml = P.BatchMatMul()
minimum = P.Minimum()
multiply = P.Mul()
transpose = P.Transpose()
get_dtype = P.DType()
alloc_status = P.NPUAllocFloatStatus()
get_status = P.NPUGetFloatStatus()
clear_status = P.NPUClearFloatStatus()
less_equal = P.LessEqual()
logical_or = P.LogicalOr()
not_equal = P.NotEqual()
select = P.Select()

_argmax = P.Argmax(axis=-1, output_type=mstype.int32)
def argmax(axis=-1, output_type=mstype.int32):
    return P.Argmax(axis=axis, output_type=output_type)



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import math
import copy
from functools import lru_cache
from turtle import position
import numpy as np
import mindspore.common.dtype as mstype
import mindspore.nn as nn
import mindspore.ops.functional as F
from mindspore.common.initializer import TruncatedNormal, initializer
from mindspore.ops import operations as P
from mindspore.ops import composite as C
import mindspore.ops as O
from mindspore.common.tensor import Tensor
from mindspore.common.parameter import Parameter


class SoftmaxCrossEntropyExpand(nn.Cell):       # pylint: disable=missing-docstring
    def __init__(self, sparse=False):
        super(SoftmaxCrossEntropyExpand, self).__init__()
        self.exp = ops.Exp()
        self.sum = ops.ReduceSum(keep_dims=True)
        self.onehot = ops.OneHot()
        self.on_value = ms.Tensor(1.0, ms.float32)
        self.off_value = ms.Tensor(0.0, ms.float32)
        self.div = ops.RealDiv()
        self.log = ops.Log()
        self.sum_cross_entropy = ops.ReduceSum(keep_dims=False)
        self.mul = ops.Mul()
        self.mul2 = ops.Mul()
        self.mean = ops.ReduceMean(keep_dims=False)
        self.sparse = sparse
        self.max = ops.ReduceMax(keep_dims=True)
        self.sub = ops.Sub()
        self.eps = ms.Tensor(1e-24, ms.float32)

    def construct(self, logit, label):      # pylint: disable=missing-docstring
        logit_max = self.max(logit, -1)
        exp = self.exp(self.sub(logit, logit_max))
        exp_sum = self.sum(exp, -1)
        softmax_result = self.div(exp, exp_sum)
        if self.sparse:
            label = self.onehot(label, ops.shape(logit)[1], self.on_value, self.off_value)

        softmax_result_log = self.log(softmax_result + self.eps)
        loss = self.sum_cross_entropy((self.mul(softmax_result_log, label)), -1)
        loss = self.mul2(ops.scalar_to_array(-1.0), loss)
        loss = self.mean(loss, -1)

        return loss


class BertPretrainingLoss(nn.Cell):
    """
    Provide bert pre-training loss.

    Args:
        config (BertConfig): The config of BertModel.

    Returns:
        Tensor, total loss.
    """

    def __init__(self, vocab_size):
        super(BertPretrainingLoss, self).__init__()
        self.vocab_size = vocab_size
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0, mstype.float32)
        self.off_value = Tensor(0.0, mstype.float32)
        self.reduce_sum = P.ReduceSum()
        self.reduce_mean = P.ReduceMean()
        self.reshape = P.Reshape()
        self.last_idx = (-1,)
        self.neg = P.Neg()
        self.cast = P.Cast()

    def construct(self, *sample):
        """Defines the computation performed.
            sample: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                     "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"] + 
                    [prediction_scores, seq_relationship_score]
        """
        masked_lm_ids = sample[5]
        masked_lm_weights = sample[6]
        prediction_scores = sample[7]
        next_sentence_labels = sample[3]
        seq_relationship_score = sample[8]
        label_ids = self.reshape(masked_lm_ids, self.last_idx)
        label_weights = self.cast(self.reshape(masked_lm_weights, self.last_idx), mstype.float32)
        one_hot_labels = self.onehot(label_ids, self.vocab_size, self.on_value, self.off_value)

        per_example_loss = self.neg(self.reduce_sum(prediction_scores * one_hot_labels, self.last_idx))
        numerator = self.reduce_sum(label_weights * per_example_loss, ())
        denominator = self.reduce_sum(label_weights, ()) + self.cast(F.tuple_to_array((1e-5,)), mstype.float32)
        masked_lm_loss = numerator / denominator

        # next_sentence_loss
        labels = self.reshape(next_sentence_labels, self.last_idx)
        if labels.max() >= 0:
            one_hot_labels = self.onehot(labels, 2, self.on_value, self.off_value)
            per_example_loss = self.neg(self.reduce_sum(
                one_hot_labels * seq_relationship_score, self.last_idx))
            next_sentence_loss = self.reduce_mean(per_example_loss, self.last_idx)
        else:
            next_sentence_loss = 0

        # total_loss
        total_loss = masked_lm_loss + next_sentence_loss

        return total_loss



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
from pyexpat import model
import shutil
from loguru import logger

import mindspore as ms
from mindspore import load_checkpoint, load_param_into_net
from mindspore.common.initializer import TruncatedNormal, initializer


def load_ckpt(model, ckpt, restore_by_prefix=True, prefix='backbone'):
    logger.info(f'Recovering last checkpoint from {ckpt}')
    params = model.get_parameters()
    # param_names = [p.name for p in params]
    param_dict = load_checkpoint(ckpt)
    # 若bert没有训练pooled out，那么backbone.bert.dense需要重新初始化
    if param_dict.get('backbone.bert.dense.weight') and param_dict['backbone.bert.dense.weight'].max() < 1e-8:
        param_dict['backbone.bert.dense.weight'].set_data(initializer(TruncatedNormal(0.02), param_dict['backbone.bert.dense.weight'].shape)) 
    if not restore_by_prefix:
        param_names = [p.name for p in params]
        unrestored = load_param_into_net(model, param_dict)
    else:
        param_names = [p.name for p in params if prefix in p.name]
        param_dict = {f'{prefix}.{k}': v for k, v in param_dict.items()}
        param_dict = {k: v for k, v in param_dict.items() if k in param_names}
        unrestored = load_param_into_net(model, param_dict)
    # unrestored = [p for p in param_names if p not in param_dict]
    missed = [p for p in param_dict.keys() if p not in param_names]
    logger.warning(f'unrestored parameters: {unrestored}\n')
    logger.warning(f'missed parameters: {missed}\n')
    return model


def save_checkpoint(state, is_best, save_dir, model_name=""):
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    filename = os.path.join(save_dir, model_name + "_ckpt.pth")
    torch.save(state, filename)
    if is_best:
        best_filename = os.path.join(save_dir, "best_ckpt.pth")
        shutil.copyfile(filename, best_filename)

if __name__ == '__main__':
    # from ..apps.models.resnet import ResidualBlock, ResNet

    # model = ResNet(ResidualBlock, 10)
    load_ckpt(None, 'output/resnet_cifar10/RESNETTask_3-6_781.ckpt')


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Local adapter"""

import os

def get_device_id():
    device_id = os.getenv('DEVICE_ID', '0')
    return int(device_id)


def get_device_num():
    device_num = os.getenv('RANK_SIZE', '1')
    return int(device_num)


def get_rank_id():
    global_rank_id = os.getenv('RANK_ID', '0')
    return int(global_rank_id)


def get_job_id():
    return "Local Job"



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import json
import copy
from mindspore import dtype

__all__=['AbsModelConfig', 'ModelConfig']

class AbsModelConfig(object):
    def __init__(self):
        pass

    @classmethod
    def from_dict(cls, json_object):
        """Constructs a `ModelConfig` from a Python dictionary of parameters."""
        config = cls()
        for key, value in json_object.items():
            if isinstance(value, dict):
                value = AbsModelConfig.from_dict(value)
            config.__dict__[key] = value
        return config

    @classmethod
    def from_json_file(cls, json_file):
        """Constructs a `ModelConfig` from a json file of parameters."""
        with open(json_file, "r", encoding='utf-8') as reader:
            text = reader.read()
        return cls.from_dict(json.loads(text))

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        def _json_default(obj):
            if isinstance(obj, AbsModelConfig):
                return obj.__dict__
        return json.dumps(self.__dict__, indent=2, sort_keys=True, default=_json_default) + "\n"
    
    def update(self, **kwargs):
        self.__dict__.update(kwargs)

    def __iter__(self):
        for k in self.__dict__.keys():
            yield k

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__.get(key)
    
    def get(self, key, default=None):
        value = self.__dict__.get(key)
        if value is None:
            value = default
        return value

class ModelConfig(AbsModelConfig):
    """Configuration class to store the configuration of a :class:`~DeBERTa.deberta.DeBERTa` model.

        Attributes:
            hidden_size (int): Size of the encoder layers and the pooler layer, default: `768`.
            num_hidden_layers (int): Number of hidden layers in the Transformer encoder, default: `12`.
            num_attention_heads (int): Number of attention heads for each attention layer in
                the Transformer encoder, default: `12`.
            intermediate_size (int): The size of the "intermediate" (i.e., feed-forward)
                layer in the Transformer encoder, default: `3072`.
            hidden_act (str): The non-linear activation function (function or string) in the
                encoder and pooler. If string, "gelu", "relu" and "swish" are supported, default: `gelu`.
            hidden_dropout_prob (float): The dropout probabilitiy for all fully connected
                layers in the embeddings, encoder, and pooler, default: `0.1`.
            attention_probs_dropout_prob (float): The dropout ratio for the attention
                probabilities, default: `0.1`.
            max_position_embeddings (int): The maximum sequence length that this model might
                ever be used with. Typically set this to something large just in case
                (e.g., 512 or 1024 or 2048), default: `512`.
            type_vocab_size (int): The vocabulary size of the `token_type_ids` passed into
                `DeBERTa` model, default: `-1`.
            initializer_range (int): The sttdev of the _normal_initializer for
                initializing all weight matrices, default: `0.02`.
            relative_attention (:obj:`bool`): Whether use relative position encoding, default: `False`.
            max_relative_positions (int): The range of relative positions [`-max_position_embeddings`, `max_position_embeddings`], default: -1, use the same value as `max_position_embeddings`. 
            padding_idx (int): The value used to pad input_ids, default: `0`.
            position_biased_input (:obj:`bool`): Whether add absolute position embedding to content embedding, default: `True`.
            pos_att_type (:obj:`str`): The type of relative position attention, it can be a combination of [`p2c`, `c2p`, `p2p`], e.g. "p2c", "p2c|c2p", "p2c|c2p|p2p"., default: "None".


    """
    def __init__(self):
        """Constructs ModelConfig.

        """
        self.architectures = [
            "BERT"
        ],
        self.attention_probs_dropout_prob = 0.1
        self.gradient_checkpointing = False
        self.hidden_act = 'gelu'
        self.hidden_dropout_prob = 0.1
        self.hidden_size = 768
        self.initializer_range = 0.02
        self.intermediate_size = 3072
        self.layer_norm_eps = 1e-05
        self.max_position_embeddings = 514
        self.num_attention_heads = 12
        self.num_hidden_layers = 12
        self.tokenizer = 'BertTokenizer'
        self.type_vocab_size = 1
        self.vocab_size = 26318
        self.seq_length = 512
        self.use_relative_positions = False
        self.compute_type = dtype.float16
        self.dtype = dtype.float32





# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from mindspore import nn
from mindspore._checkparam import Validator as validator
from mindspore._checkparam import Rel
from mindspore.common import dtype as mstype
from mindspore.nn.wrap.cell_wrapper import _TrainPipelineAccuStepCell
from mindspore.nn.wrap.loss_scale import _TrainPipelineWithLossScaleCell
from mindspore.ops import functional as F
from mindspore.parallel._utils import _get_pipeline_stages
from mindspore.train.loss_scale_manager import DynamicLossScaleManager, LossScaleManager
from mindspore import boost
from mindspore import context


class _OutputTo16(nn.Cell):
    "Wrap cell for amp. Cast network output back to float16"

    def __init__(self, op):
        super(_OutputTo16, self).__init__(auto_prefix=False)
        self._op = op

    def construct(self, x):
        return F.cast(self._op(x), mstype.float16)


def _do_keep_batchnorm_fp32(network):
    """Do keep batchnorm fp32."""
    cells = network.name_cells()
    change = False
    for name in cells:
        subcell = cells[name]
        if subcell == network:
            continue
        elif isinstance(subcell, (nn.BatchNorm2d, nn.BatchNorm1d)):
            network._cells[name] = _OutputTo16(subcell.to_float(mstype.float32))
            change = True
        elif isinstance(subcell, nn.LossBase):
            network._cells[name] = subcell.to_float(mstype.float32)
            change = True
        else:
            _do_keep_batchnorm_fp32(subcell)
    if isinstance(network, nn.SequentialCell) and change:
        network.cell_list = list(network.cells())


_config_level = {
    "O0": {
        "keep_batchnorm_fp32": False,
        "cast_model_type": mstype.float32,
        "loss_scale_manager": None},
    "O2": {
        "keep_batchnorm_fp32": True,
        "cast_model_type": mstype.float16,
        "loss_scale_manager": DynamicLossScaleManager()},
    "O3": {
        "keep_batchnorm_fp32": False,
        "cast_model_type": mstype.float16,
        "loss_scale_manager": None}}


def _check_kwargs(key_words):
    """Check kwargs."""
    for arg in key_words:
        if arg not in ['cast_model_type', 'keep_batchnorm_fp32', 'loss_scale_manager']:
            raise ValueError(f"Unsupported arg '{arg}'")

    if 'cast_model_type' in key_words:
        validator.check_type_name('cast_model_type', key_words['cast_model_type'],
                                  [mstype.float16, mstype.float32], None)
    if 'keep_batchnorm_fp32' in key_words:
        validator.check_value_type('keep_batchnorm_fp32', key_words['keep_batchnorm_fp32'], bool)
    if 'loss_scale_manager' in key_words:
        loss_scale_manager = key_words['loss_scale_manager']
        if loss_scale_manager:
            validator.check_value_type('loss_scale_manager', loss_scale_manager, LossScaleManager)


def _check_level(level, boost_level):
    """Check level."""
    if not isinstance(level, str):
        raise TypeError("The argument `level` must be a string in ['O0', 'O2', 'O3', 'auto'], \
                         but got type {}.".format(type(level)))
    validator.check('level', level, "", ['O0', 'O2', 'O3', 'auto'], Rel.IN)
    validator.check('boost_level', boost_level, "", ['O0', 'O1', 'O2'], Rel.IN)

    if level == "auto":
        device_target = context.get_context('device_target')
        if device_target == "GPU":
            level = "O2"
        elif device_target == "Ascend":
            level = "O3"
        else:
            raise ValueError("Level `auto` only support when `device_target` is GPU or Ascend.")

    enable_boost = False
    if boost_level in ["O1", "O2"]:
        enable_boost = True

    return level, enable_boost


def _add_loss_network(network, loss_fn, cast_model_type):
    """Add loss network."""

    # class WithLossCell(nn.Cell):
    #     "Wrap loss for amp. Cast network output back to float32"

    #     def __init__(self, backbone, loss_fn):
    #         super(WithLossCell, self).__init__(auto_prefix=False)
    #         self._backbone = backbone
    #         self._loss_fn = loss_fn
    #         # self.return_all = True

    #     def construct(self, *args):
    #         data = args[0:-1]
    #         label = args[-1]
    #         out = self._backbone(data)
    #         label = F.mixed_precision_cast(mstype.float32, label)
    #         out = F.mixed_precision_cast(mstype.float32, out)
    #         loss = self._loss_fn(out, label)
    #         return loss

    from ..modeling.modeling_adapter import NetworkWithLoss
    validator.check_value_type('loss_fn', loss_fn, nn.Cell)
    if cast_model_type == mstype.float16:
        network = NetworkWithLoss(network, loss_fn, fp16=True)
    else:
        network = NetworkWithLoss(network, loss_fn)
    return network


def build_train_network(network, loss_fn=None, level='O0', boost_level='O0', **kwargs):
    """
    Build the mixed precision training cell automatically.

    Args:
        network (Cell): Definition of the network.
        loss_fn (Union[None, Cell]): Define the loss function. If None, the `network` should have the loss inside.
            Default: None.
        optimizer (Optimizer): Define the optimizer to update the Parameter.
        level (str): Supports ["O0", "O2", "O3", "auto"]. Default: "O0".

            - "O0": Do not change.
            - "O2": Cast network to float16, keep batchnorm and `loss_fn` (if set) run in float32,
              using dynamic loss scale.
            - "O3": Cast network to float16, with additional property `keep_batchnorm_fp32=False` .
            - auto: Set to level to recommended level in different devices. Set level to "O2" on GPU, Set
              level to "O3" Ascend. The recommended level is chosen by the export experience, not applicable to all
              scenarios. User should specify the level for special network.

            "O2" is recommended on GPU, "O3" is recommended on Ascend. Property of `keep_batchnorm_fp32`,
            `cast_model_type` and `loss_scale_manager` determined by `level` setting may be overwritten by settings in
            `kwargs`.

        boost_level (str): Option for argument `level` in `mindspore.boost` , level for boost mode
            training. Supports ["O0", "O1", "O2"]. Default: "O0".

            - "O0": Do not change.
            - "O1": Enable the boost mode, the performance is improved by about 20%, and
              the accuracy is the same as the original accuracy.
            - "O2": Enable the boost mode, the performance is improved by about 30%, and
              the accuracy is reduced by less than 3%.

            If "O1" or "O2" mode is set, the boost related library will take effect automatically.

        cast_model_type (:class:`mindspore.dtype`): Supports `mstype.float16` or `mstype.float32` . If set, the
            network will be casted to `cast_model_type` ( `mstype.float16` or `mstype.float32` ), but not to be casted
            to the type determined by `level` setting.
        keep_batchnorm_fp32 (bool): Keep Batchnorm run in `float32` when the network is set to cast to `float16` . If
            set, the `level` setting will take no effect on this property.
        loss_scale_manager (Union[None, LossScaleManager]): If not None, must be subclass of
            :class:`mindspore.LossScaleManager` for scaling the loss. If set, the `level` setting will take no effect
            on this property.

    Raises:
        ValueError: If device is CPU, property `loss_scale_manager` is not `None` or `FixedLossScaleManager`
            (with property `drop_overflow_update=False` ).
    """
    validator.check_value_type('network', network, nn.Cell)
    # validator.check_value_type('optimizer', optimizer, (nn.Optimizer, boost.FreezeOpt,
    #                                                     nn.AdaSumByGradWrapCell, nn.AdaSumByDeltaWeightWrapCell))

    level, enable_boost = _check_level(level, boost_level)

    _check_kwargs(kwargs)
    config = dict(_config_level[level], **kwargs)

    if config["cast_model_type"] == mstype.float16:
        network.to_float(mstype.float16)

        if config["keep_batchnorm_fp32"]:
            _do_keep_batchnorm_fp32(network)

    if loss_fn:
        network = _add_loss_network(network, loss_fn, config["cast_model_type"])

    # loss_scale = 1.0
    # if config["loss_scale_manager"] is not None:
    #     loss_scale_manager = config["loss_scale_manager"]
    #     loss_scale = loss_scale_manager.get_loss_scale()
    #     update_cell = loss_scale_manager.get_update_cell()
    #     if update_cell is not None:
    #         # only cpu not support `TrainOneStepWithLossScaleCell` for control flow.
    #         if not context.get_context("enable_ge") and context.get_context("device_target") == "CPU":
    #             raise ValueError("Only `loss_scale_manager=None` or "
    #                              "`loss_scale_manager=FixedLossScaleManager(drop_overflow_update=False)`"
    #                              "are supported on device `CPU`. ")
    #         if _get_pipeline_stages() > 1:
    #             network = _TrainPipelineWithLossScaleCell(network, optimizer,
    #                                                       scale_sense=update_cell).set_train()
    #         elif enable_boost:
    #             network = boost.BoostTrainOneStepWithLossScaleCell(network, optimizer,
    #                                                                scale_sense=update_cell).set_train()
    #         else:
    #             network = nn.TrainOneStepWithLossScaleCell(network, optimizer,
    #                                                        scale_sense=update_cell).set_train()
    #         return network
    # if _get_pipeline_stages() > 1:
    #     network = _TrainPipelineAccuStepCell(network, optimizer).set_train()
    # elif enable_boost:
        # network = boost.BoostTrainOneStepCell(network, optimizer, loss_scale).set_train()
    # else:
    #     network = nn.TrainOneStepCell(network, optimizer, loss_scale).set_train()
    return network



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import OrderedDict
import numpy as np
import os
import random
from shutil import copyfile
from loguru import logger

from MindsporeTrainer.data import ExampleInstance, ExampleSet, _truncate_segments
from MindsporeTrainer.data.example import *
from MindsporeTrainer.task import TransformerTask
from MindsporeTrainer.task  import register_task
from MindsporeTrainer.utils.metrics import *
from MindsporeTrainer.utils.metrics import BertMetric
from MindsporeTrainer.utils.masker import NGramMaskGenerator
from MindsporeTrainer.data.dynamic_dataset import create_dynamic_dataset
from MindsporeTrainer.modeling.tokenizers import BertTokenizer

@register_task(name="DEBERTA", desc="Basic DEBERTA task")
class DEBERTATask(TransformerTask):
    def __init__(self, args, **kwargs):
        super().__init__(args, **kwargs)
        self.max_seq_len = 512
        self.model_config = 'data/model_config.json'
        self.vocab_type = 'BERT'
        self.vocab_path = 'data/vocab.txt'
        self.metric = 'bert'
        self.main_metric = 'perplexity'
        self.optimizer_name = 'Lamb'
        self.tokenizer = BertTokenizer(self.vocab_path)

    def train_data(self, **kwargs):
        data_path = os.path.join(self.data_dir, 'daizhige.pkl')
        data = self.load_data(data_path)
        # data = ExampleSet(data)
        output_columns = ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                                    "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
        return create_dynamic_dataset(data, [self.get_feature_fn()], 
                                      batch_size=self.train_batch_size,
                                      output_columns=output_columns, 
                                      repeat=self.num_train_epochs,
                                      num_workers=self.data_workers,
                                      num_shards=self.rank_size, 
                                      shard_id=self.rank_id)

    def eval_data(self, **kwargs):
        data_path = os.path.join(self.data_dir, 'eval.txt')
        data = self.load_txt_data(data_path)

        output_columns = ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                                    "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
        return create_dynamic_dataset(data, self.get_feature_fn(), 
                                      batch_size=self.eval_batch_size,
                                      output_columns=output_columns, 
                                      repeat=1,
                                      num_workers=self.data_workers, 
                                      num_shards=self.rank_size, 
                                      shard_id=self.rank_id)

    def get_metrics(self, **kwargs):
        """Calcuate metrics based on prediction results"""
        return OrderedDict(
            bert = BertMetric(self.eval_batch_size),
            )

    def get_eval_fn(self, **kwargs):
        data = kwargs.get('data')
        if data is None:
            data = self.eval_data(**kwargs)
        def run_eval(model, name, prefix):
            '''
            args: 
                model: Model instance
                name: evaluate name
                prefix: prefix of file
            return:
                float, main metric of this task, used to save best metric model
            '''
            res = model.eval(data, dataset_sink_mode=False)
            res = res['bert']
            main_metric = res[self.main_metric]
            if self.rank_id == 0:
                output=os.path.join(self.output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
                metric_str = '\n'.join([f'{k}: {v:.4f}' for k, v in res.items()])
                metric_str = metric_str + '\n'
                logger.info("====================================")
                logger.info("evaluate result:\n")
                logger.info(metric_str)
                logger.info("====================================")

                with open(output, 'w', encoding='utf-8') as fs:
                    fs.write(f"metrics:\n{metric_str}\n")
            return main_metric
        return run_eval


    def get_feature_fn(self, ext_params=None, rng=None, **kwargs):
        tokenizer = self.tokenizer
        mask_generator = NGramMaskGenerator(tokenizer)
        def example_to_feature(*example):
            '''
            sample: text, label
            return: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                    masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
            '''
            r = rng
            if r is None:
                r = random
            max_num_tokens = self.max_seq_len - 2
            segments = example[0].tolist().decode('utf-8')
            _tokens = ['[CLS]'] + tokenizer.tokenize(segments) + ['[SEP]']
            if mask_generator:
                    tokens, lm_labels = mask_generator.mask_tokens(_tokens, r)
            token_ids = tokenizer.convert_tokens_to_ids(tokens)
            pad_id = tokenizer._convert_token_to_id(tokenizer.pad_token)
            seq_len = len(token_ids)
            pad_num = self.max_seq_len - seq_len
            token_ids = token_ids + [pad_id] * pad_num
            input_mask = [1] * seq_len + [0] * pad_num
            token_type_id = [0] * self.max_seq_len
            next_sentence_labels = -1
            masked_lm_positions = [i for i, lbl in enumerate(lm_labels) if lbl != 0]
            masked_lm_ids = [lbl for i, lbl in enumerate(lm_labels) if lbl != 0]
            if len(masked_lm_ids) < mask_generator.max_preds_per_seq:
                masked_lm_positions += [0] * (mask_generator.max_preds_per_seq - len(masked_lm_ids))
                masked_lm_ids += [-1] * (mask_generator.max_preds_per_seq - len(masked_lm_ids))
            masked_lm_weights = [1] * len(masked_lm_ids)
            features = (token_ids,
                        input_mask,
                        token_type_id,
                        next_sentence_labels,
                        masked_lm_positions,
                        masked_lm_ids,
                        masked_lm_weights)

            features = [np.array(f, dtype=np.int32) for f in features]
            return tuple(features)

        return example_to_feature

    def load_txt_data(self, path, type=None):
        examples = []
        with open(path, encoding='utf-8') as fs:
                current_chunk = ''
                current_length = 0
                for l in fs:
                    l = l.replace('\t', '')
                    if self.args.debug:
                        if len(examples) >= 3000:
                            return examples
                    if type == 'JW':
                        if random.random() > 0.5:
                            continue
                    if '\n' in l:
                        l_list = l.split('\n')
                    else:
                        l_list = [l]
                    for l in l_list:
                        if len(l) == 0:
                            continue
                        current_chunk += l
                        if len(current_chunk) >= self.max_seq_len:
                            example = ExampleInstance(segments=[current_chunk[: self.max_seq_len - 2]])
                            if len(example.segments[0]) < 510:
                                print(len(example.segments[0]))
                            examples.append(example)
                            current_chunk = current_chunk[self.max_seq_len - 2: ]
        return ExampleSet(examples)

    def load_data(self, path, type=None):
        import pickle
        examples = []
        with open(path, 'rb') as fs:
            data = pickle.load(fs)
        for cat, files in data.items():
            for path, lines in files.items():
                current_chunk = ''
                current_length = 0
                for l in lines:
                    if self.args.debug:
                        if len(examples) >= 30000:
                            return ExampleSet(examples)
                    if len(l) == 0:
                        continue
                    current_chunk += l
                    if len(current_chunk) >= self.max_seq_len:
                        if current_chunk[0] in ['O', '。', '？', '！', '，', '、', '；', '：', '|']:
                            current_chunk = current_chunk[1:]
                        example = ExampleInstance(segments=[current_chunk[: self.max_seq_len - 2]])
                        if len(example.segments[0]) < 510:
                            print(len(example.segments[0]))
                        # examples.append(np.array([current_chunk[: max_seq_len - 2]]))
                        examples.append(example)
                        current_chunk = current_chunk[self.max_seq_len - 2: ]
        return ExampleSet(examples)


    def get_model(self):
        from MindsporeTrainer.modeling.models import BertPreTraining, DebertaPreTraining
        from MindsporeTrainer import build_transformer_model
        if self.args.fp16:
            compute_type = mstype.float16
        else:
            compute_type = mstype.float32
        model, config = build_transformer_model(self.model_config, 
                                                model=DebertaPreTraining, 
                                                compute_type=compute_type, 
                                                padding_idx=self.tokenizer._convert_token_to_id(self.tokenizer.pad_token))
        with open(os.path.join(self.output_dir, 'config.json'), 'w', encoding='utf-8') as f:
            f.write(config.to_json_string())
        copyfile(self.vocab_path, os.path.join(self.output_dir, 'vocab.txt'))
        return model
        # return partial_class

    def get_loss(self, *args, **kwargs):
        from MindsporeTrainer.modeling.loss import BertPretrainingLoss
        return BertPretrainingLoss(self.tokenizer.vocab_size)

    def get_eval_head(self, *args, **kwargs):
        from MindsporeTrainer.modeling.layers import BertEvalHead
        return BertEvalHead(self.tokenizer.vocab_size)

    def get_opt_fn(self, *args, **kwargs):
        return None





# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from mindspore import context
from mindspore.nn.optim import Lamb, Momentum, AdamWeightDecay, SGD
from .utils import LearningRate

from .args import get_args
from .adam import *


def get_optimizer(args_opt, network, opt_name='AdamWeightDecay'):
    """get ernie optimizer, support Lamb, Momentum, AdamWeightDecay."""
    params = network.trainable_params()
    decay_filter = lambda x: 'layernorm' not in x.name.lower() and 'bias' not in x.name.lower()
    decay_params = list(filter(decay_filter, params))
    other_params = list(filter(lambda x: not decay_filter(x), params))
    group_params = [{'params': decay_params, 'weight_decay': 1e-2},
                    {'params': other_params, 'weight_decay': 0.0},
                    {'order_params': params}]
    opt_overflow = False
    if opt_name == 'Lamb':
        lr_schedule = LearningRate(learning_rate=args_opt.learning_rate,
                                    end_learning_rate=0.0,
                                    warmup_steps=int(args_opt.warmup * args_opt.train_steps),
                                    decay_steps=args_opt.train_steps,
                                    power=2.0)

        optimizer = Lamb(group_params, learning_rate=lr_schedule, eps=1e-8)
    elif opt_name == 'Momentum':
        optimizer = Momentum(filter(lambda x: x.requires_grad, network.get_parameters()), learning_rate=args_opt.learning_rate,  #network.trainable_params()
                             momentum=0.9)
    elif opt_name == 'AdamWeightDecay':
        lr_schedule = LearningRate(learning_rate=args_opt.learning_rate,
                                    end_learning_rate=0.0,
                                    warmup_steps=int(args_opt.warmup * args_opt.train_steps),
                                    decay_steps=args_opt.train_steps,
                                    power=5.0)

        if args_opt.enable_lossscale == "true" and args_opt.device_target == 'GPU':
            optimizer = AdamWeightDecayX(group_params, learning_rate=lr_schedule, eps=1e-6)
            opt_overflow = True
        elif context.get_context("mode") == context.PYNATIVE_MODE and args_opt.device_target == 'GPU':
            optimizer = AdamWeightDecayOp(group_params, learning_rate=lr_schedule, eps=1e-6)
        else:
            optimizer = AdamWeightDecay(group_params, learning_rate=lr_schedule, eps=1e-6)
            opt_overflow = True
    else:
        raise ValueError("Don't support optimizer {}, only support [Lamb, Momentum, AdamWeightDecay, Thor]".
                         format(opt_name))
    return optimizer, opt_overflow



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import OrderedDict

import mindspore.dataset as ds
import mindspore.dataset.vision as vision
import mindspore.dataset.transforms as transforms
from mindspore.communication import get_rank, get_group_size
from mindspore.nn import SoftmaxCrossEntropyWithLogits
import mindspore as ms

from MindsporeTrainer.task import Task
from MindsporeTrainer.task  import register_task
from MindsporeTrainer.utils.metrics import *

@register_task(name="RESNET", desc="a task demo for resnet on cifar10")
class RESNETTask(Task):
    def __init__(self, args, **kwargs):
        super().__init__(args, **kwargs)
        self.args = args
        self.data_dir = args.data_dir
        self.metric = {'acc': MSAucuracy()}
        self.main_metric = 'acc'
        self.optimizer_name = 'Momentum'

    def train_data(self, **kwargs):
        return self.create_dataset(usage='train', batch_size=self.train_batch_size, rank_id=self.rank_id, rank_size=self.rank_size)

    def eval_data(self, **kwargs):
        return self.create_dataset(usage='test', batch_size=self.predict_batch_size, rank_id=self.rank_id, rank_size=self.rank_size)

    def test_data(self, **kwargs):
        return self.create_dataset(usage='test', batch_size=self.eval_batch_size, rank_id=self.rank_id, rank_size=self.rank_size)

    def get_metrics(self, **kwargs):
        """Calcuate metrics based on prediction results"""
        return {'acc': MSAucuracy()}

    def get_eval_fn(self, **kwargs):
        '''
        we use a default eval function
        '''
        return None

    def get_feature_fn(self, **kwargs):
        '''
        we do not need a feature function here
        '''
        def _example_to_feature(**kwargs):
            return None
        return _example_to_feature

    def get_labels(self):
        """labels of this task."""
        return [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    def create_dataset(self, usage='train', repeat_num=1, batch_size=32, rank_id=0, rank_size=1):
        '''
        build the dataset
        '''
        resize_height = 224
        resize_width = 224
        rescale = 1.0 / 255.0
        shift = 0.0

        data_set = ds.Cifar10Dataset(self.data_dir, usage=usage, num_shards=rank_size, shard_id=rank_id, num_parallel_workers=4)

        # define map operations
        random_crop_op = vision.c_transforms.RandomCrop((32, 32), (4, 4, 4, 4))
        random_horizontal_op = vision.c_transforms.RandomHorizontalFlip()
        resize_op = vision.c_transforms.Resize((resize_height, resize_width))
        rescale_op = vision.c_transforms.Rescale(rescale, shift)
        normalize_op = vision.c_transforms.Normalize((0.4465, 0.4822, 0.4914), (0.2010, 0.1994, 0.2023))
        changeswap_op = vision.c_transforms.HWC2CHW()
        type_cast_op = transforms.c_transforms.TypeCast(ms.int32)

        c_trans = [random_crop_op, random_horizontal_op]
        c_trans += [resize_op, rescale_op, normalize_op, changeswap_op]

        # apply map operations on images
        data_set = data_set.map(operations=type_cast_op, input_columns="label")
        data_set = data_set.map(operations=c_trans, input_columns="image")

        # apply shuffle operations
        data_set = data_set.shuffle(buffer_size=10)

        # apply batch operations
        data_set = data_set.batch(batch_size=batch_size, drop_remainder=True)

        # apply repeat operations
        data_set = data_set.repeat(repeat_num)

        return data_set

    def get_model(self):
        '''
        build the model
        '''
        from MindsporeTrainer.apps.models.resnet import ResidualBlock, ResNet

        return ResNet(ResidualBlock, len(self.get_labels()))

    def get_loss(self, *args, **kwargs):
        '''
        apply cross entropy loss
        '''
        return SoftmaxCrossEntropyWithLogits(sparse=True, reduction='mean')

    def get_eval_head(self, *args, **kwargs):
        from MindsporeTrainer.modeling.layers import ClsEvalHead
        return ClsEvalHead(len(self.get_labels()))

    def get_opt_fn(self, *args, **kwargs):
        '''
        we use a default optimizer
        '''
        return None


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

class BertTokenizer(object):
    def __init__(self, **kwargs,) -> None:
        pass

class Vocab(object):
    def __init__(self, **kwargs,) -> None:
        pass

    @staticmethod
    def from_dict(vocab):
        return None


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
import mindspore.nn as nn
import mindspore.ops as ops
import mindspore.dataset as ds
import mindspore.dataset.vision as vision
import mindspore.dataset.transforms as transforms
from mindspore.communication import init, get_rank, get_group_size
from mindspore.nn import Momentum
from mindspore.nn.metrics import Metric
from mindspore.ops.composite import GradOperation
import mindspore as ms
from MindsporeTrainer.apps.models.resnet import ResidualBlock, ResNet, resnet50
from mindspore.parallel._ps_context import _is_role_pserver, _is_role_sched

ms.context.set_context(mode=ms.context.PYNATIVE_MODE, device_target='Ascend', device_id=0)
# init("nccl")

def create_dataset(data_path, repeat_num=1, batch_size=32, rank_id=0, rank_size=1):     # pylint: disable=missing-docstring
    resize_height = 224
    resize_width = 224
    rescale = 1.0 / 255.0
    shift = 0.0

    # get rank_id and rank_size
    # rank_id = get_rank()
    # rank_size = get_group_size()
    indices = list(range(128))
    sampler = ds.SubsetRandomSampler(indices=indices)
    data_set = ds.Cifar10Dataset(data_path)

    # define map operations
    random_crop_op = vision.c_transforms.RandomCrop((32, 32), (4, 4, 4, 4))
    random_horizontal_op = vision.c_transforms.RandomHorizontalFlip()
    resize_op = vision.c_transforms.Resize((resize_height, resize_width))
    rescale_op = vision.c_transforms.Rescale(rescale, shift)
    normalize_op = vision.c_transforms.Normalize((0.4465, 0.4822, 0.4914), (0.2010, 0.1994, 0.2023))
    changeswap_op = vision.c_transforms.HWC2CHW()
    type_cast_op = transforms.c_transforms.TypeCast(ms.int32)

    c_trans = [random_crop_op, random_horizontal_op]
    c_trans += [resize_op, rescale_op, normalize_op, changeswap_op]

    # apply map operations on images
    data_set = data_set.map(operations=type_cast_op, input_columns="label")
    data_set = data_set.map(operations=c_trans, input_columns="image")

    # apply shuffle operations
    data_set = data_set.shuffle(buffer_size=10)

    # apply batch operations
    data_set = data_set.batch(batch_size=batch_size, drop_remainder=True)

    # apply repeat operations
    data_set = data_set.repeat(repeat_num)

    return data_set


class SoftmaxCrossEntropyExpand(nn.Cell):       # pylint: disable=missing-docstring
    def __init__(self, sparse=False):
        super(SoftmaxCrossEntropyExpand, self).__init__()
        self.exp = ops.Exp()
        self.sum = ops.ReduceSum(keep_dims=True)
        self.onehot = ops.OneHot()
        self.on_value = ms.Tensor(1.0, ms.float32)
        self.off_value = ms.Tensor(0.0, ms.float32)
        self.div = ops.RealDiv()
        self.log = ops.Log()
        self.sum_cross_entropy = ops.ReduceSum(keep_dims=False)
        self.mul = ops.Mul()
        self.mul2 = ops.Mul()
        self.mean = ops.ReduceMean(keep_dims=False)
        self.sparse = sparse
        self.max = ops.ReduceMax(keep_dims=True)
        self.sub = ops.Sub()
        self.eps = ms.Tensor(1e-24, ms.float32)

    def construct(self, logit, label):      # pylint: disable=missing-docstring
        logit_max = self.max(logit, -1)
        exp = self.exp(self.sub(logit, logit_max))
        exp_sum = self.sum(exp, -1)
        softmax_result = self.div(exp, exp_sum)
        if self.sparse:
            label = self.onehot(label, ops.shape(logit)[1], self.on_value, self.off_value)

        softmax_result_log = self.log(softmax_result + self.eps)
        loss = self.sum_cross_entropy((self.mul(softmax_result_log, label)), -1)
        loss = self.mul2(ops.scalar_to_array(-1.0), loss)
        loss = self.mean(loss, -1)

        return loss

class BertMetric(Metric):
    """
    The metric of bert network.
    Args:
        batch_size (int): The batchsize of each device.
    """
    def __init__(self):
        super(BertMetric, self).__init__()
        self.clear()

    def clear(self):
        self.logits = None
        self.labels = None

    def update(self, *inputs):
        logits, labels = inputs
        if self.logits is None:
            self.logits = logits.argmax(-1)
            self.labels = labels
        else:
            self.logits = ms.ops.Concat()([self.logits, logits.argmax(-1)])
            self.labels = ms.ops.Concat()([self.labels, labels])

    def eval(self):
        acc = self.logits == self.labels
        acc = acc.sum()
        total = self.labels.size
        acc = acc.astype('float') / total
        return acc

import time
class StepTimeMonitor(ms.train.callback.Callback):
    def __init__(self, data_size=None):
        super(StepTimeMonitor, self).__init__()
        self.data_size = data_size
        self.step_time = 0
        self.steps = 0
        self.results = []
    
    def step_begin(self, run_context):
        """
        Called before each step beginning.

        Args:
            run_context (RunContext): Include some information of the model.
        """
        self.step_time = time.time()

    def step_end(self, run_context):
        """
        Called after each step finished.

        Args:
            run_context (RunContext): Include some information of the model.
        """
        self.steps += 1
        self.results.append(run_context.original_args().net_outputs)
        print(f'step time: {time.time() - self.step_time}')


def test_train_cifar(epoch_size=10):        # pylint: disable=missing-docstring
    # ms.set_auto_parallel_context(parallel_mode=ms.ParallelMode.AUTO_PARALLEL, gradients_mean=True)
    loss_cb = ms.train.callback.LossMonitor()
    time_cb = ms.train.callback.TimeMonitor(data_size=1)
    batch_size = 256
    num_classes = 10
    data_path = 'data/tasks/cifar-10-batches-bin'
    dataset = create_dataset(data_path, batch_size=batch_size)

    net = resnet50(32,10).set_train()# ResNet(ResidualBlock, num_classes, batch_size)
    net = ResNet(ResidualBlock, num_classes, batch_size)
    scale_factor = 4
    scale_window = 3000
    # loss = nn.SoftmaxCrossEntropyWithLogits(sparse=True, reduction='mean')
    loss = SoftmaxCrossEntropyExpand(sparse=True)
    opt = Momentum(filter(lambda x: x.requires_grad, net.get_parameters()), 0.01, 0.9)
    loss_scale_manager = ms.DynamicLossScaleManager(scale_factor, scale_window)
    net = ms.amp.build_train_network(net, opt, level='O3', loss_scale_manager=loss_scale_manager, loss_fn=loss)

    
    grad_op = GradOperation(get_by_list=True, sens_param=True)
    batch = next(dataset.create_tuple_iterator())
    weigths = [w for w in net.get_parameters()]
    logits = net(*batch)
    grads = grad_op(net, weigths)(*batch, ms.Tensor(10, dtype=ms.dtype.float32))
    model = ms.Model(net, loss_fn=loss, optimizer=opt, amp_level='O0', metrics={'acc'})
    model.train(1, dataset, callbacks=[loss_cb, StepTimeMonitor()], dataset_sink_mode=False)
    # m = StepTimeMonitor()
    # metrics = model.eval(dataset, callbacks=[m])
    # print(metrics)

if __name__ == '__main__':
    # _is_role_sched()
    test_train_cifar()


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from glob import glob
import os
import importlib
import pdb
import sys
from loguru import logger
from .task import Task

__all__ = ['load_tasks', 'register_task', 'get_task']
tasks = {}

def register_task(name=None, desc=None):
  global tasks
  def register_task_x(cls):
    _name = name
    if _name is None:
      _name = cls.__name__

    _desc = desc
    if _desc is None:
      _desc = _name

    _name = _name.lower()
    if _name in tasks:
      logger.warning(f'{_name} already registered in the registry: {tasks[_name]}.')
    assert issubclass(cls, Task), f'Registered class must be a subclass of Task.'
    tasks[_name] = cls
    cls._meta = {
        'name': _name,
        'desc': _desc}
    return cls
  
  if type(name)==type:
    cls = name
    name = None
    return register_task_x(cls)
  return register_task_x

def load_tasks(task_dir=None, task_path=None):
  if task_path:
    if '.py' in task_path and os.path.exists(task_path):
      task_path = task_path.replace(os.getcwd(), '').replace('.py', '')
      m = task_path.replace(os.path.sep, '.')
      importlib.import_module(f'{m}')
      return
  script_dir = os.path.dirname(os.path.abspath(__file__))
  base_dir = script_dir.replace('/task', '')
  script_dir = base_dir + '/apps/tasks'
  sys.path.append(script_dir)
  sys_tasks = glob(os.path.join(script_dir, "*.py"))
  for t in sys_tasks:
    m = os.path.splitext(os.path.basename(t))[0]
    if not m.startswith('_'):
      importlib.import_module(f'apps.tasks.{m}')

  if task_dir:
    assert os.path.exists(task_dir), f"{task_dir} must be a valid directory."
    customer_tasks = glob(os.path.join(task_dir, "*.py"))
    sys.path.append(task_dir)
    for t in customer_tasks:
      m = os.path.splitext(os.path.basename(t))[0]
      if not m.startswith('_'):
        importlib.import_module(f'{m}')

def get_task(name=None):
  global tasks
  if name is None:
    return tasks
  else:
    if '.py' in name and os.path.exists(name):
      return list(tasks.values())[0]
    else:
      return tasks[name.lower()]



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from ast import Not
import time, os
import mindspore as ms
import numpy as np
from loguru import logger
from mindspore import Tensor
from mindspore.train.callback import Callback, ModelCheckpoint, _set_cur_net
import mindspore.context as context
from mindspore.train.serialization import save_checkpoint, _save_graph
from mindspore.context import ParallelMode
from mindspore.ops import operations as P
from mindspore import load_checkpoint, load_param_into_net

from MindsporeTrainer.modeling.modeling_adapter import ModelForEval, AllGather

class StepTimeMonitor(Callback):
    def __init__(self, data_size=None):
        super(StepTimeMonitor, self).__init__()
        self.data_size = data_size
        self.step_time = 0
    
    def step_begin(self, run_context):
        """
        Called before each step beginning.

        Args:
            run_context (RunContext): Include some information of the model.
        """
        self.step_time = time.time()

    def step_end(self, run_context):
        """
        Called after each step finished.

        Args:
            run_context (RunContext): Include some information of the model.
        """
        print(f'step time: {time.time() - self.step_time}')

class LossMoniter(ms.train.callback.LossMonitor):
    def step_end(self, run_context):
        """
        Print training loss at the end of step.

        Args:
            run_context (RunContext): Include some information of the model.
        """
        cb_params = run_context.original_args()
        if not 'cur_epoch_num' in cb_params:
            cb_params.cur_epoch_num = 0

        loss = cb_params.net_outputs

        if isinstance(loss, (tuple, list)):
            if isinstance(loss[0], Tensor) and isinstance(loss[0].asnumpy(), np.ndarray):
                loss = loss[0]

        if isinstance(loss, Tensor) and isinstance(loss.asnumpy(), np.ndarray):
            loss = float(np.mean(loss.asnumpy()))

        cur_step_in_epoch = (cb_params.cur_step_num - 1) % cb_params.batch_num + 1

        if isinstance(loss, float) and (np.isnan(loss) or np.isinf(loss)):
            raise ValueError("epoch: {} step: {}. Invalid loss, terminating training.".format(
                cb_params.cur_epoch_num, cur_step_in_epoch))

        # #In disaster recovery scenario, the cb_params.cur_step_num may be rollback to previous step
        # # and be less than self._last_print_time, so self._last_print_time need to be updated.
        # if self._per_print_times != 0 and (cb_params.cur_step_num <= self._last_print_time):
        #     while cb_params.cur_step_num <= self._last_print_time:
        #         self._last_print_time -=\
        #             max(self._per_print_times, cb_params.batch_num if cb_params.dataset_sink_mode else 1)

        if cb_params.cur_step_num % self._per_print_times ==0 :
            # self._last_print_time = cb_params.cur_step_num
            logger.info("Evaluating: steps: %s, loss is %s" % (cur_step_in_epoch, loss))

class EvalCallBack(Callback):
    """
    Evaluate after a certain amount of training samples.
    Args:
        model (Model): The network model.
        eval_ds (Dataset): The eval dataset.
        global_batch (int): The batchsize of the sum of all devices.
        eval_samples (int): The number of eval interval samples.
    """
    def __init__(self, model, trainer_state, eval_ds, global_batch, eval_fn, train_steps,
                    eval_steps=0, eval_samples=None, rank=0):
        super(EvalCallBack, self).__init__()
        self.model = model
        self.eval_ds = eval_ds
        self.global_batch = global_batch
        self.eval_samples = eval_samples
        self.last_eval_step = 0
        self.eval_steps = eval_steps
        self.trainer_state = trainer_state
        self.eval_fn = eval_fn
        self.rank = rank
        self._last_triggered_step = 0
        self.train_steps = train_steps


    def step_end(self, run_context):
        if self.eval_steps == 0:
            return
        else:
            if (run_context.original_args().cur_step_num  >= self._last_triggered_step + self.eval_steps):
                self.evaluate(run_context)

    def epoch_end(self, run_context):
        """
        Evaluate after training a certain number of samples.
        """
        if self.eval_steps > 0:
            return
        else:
            self.evaluate(run_context)
    
    def evaluate(self, run_context):
        cb_params = run_context.original_args()
        self._last_triggered_step = cb_params.cur_step_num 

        metric = self.eval_fn(self.model, 'eval', cb_params.cur_step_num)
        if self.trainer_state.neg_metric:
            metric = -metric
        if metric > self.trainer_state.best_metric:
            self.trainer_state.best_metric = metric
            self.trainer_state.best_steps = cb_params.cur_step_num

    def end(self, run_context):
        """
        Called once after network training.

        Args:
            run_context (RunContext): Include some information of the model.
        """
        self.evaluate(run_context)


class SummaryCallback(Callback):
    def __init__(self, summary_dir):
        self._summary_dir = summary_dir

    def __enter__(self):
        # init you summary record in here, when the train script run, it will be inited before training
        self.summary_record =  ms.train.summary.SummaryRecord(self._summary_dir)
        return self

    def __exit__(self, *exc_args):
        # Note: you must close the summary record, it will release the process pool resource
        # else your training script will not exit from training.
        self.summary_record.close()

    def step_end(self, run_context):
        cb_params = run_context.original_args()

        # create a confusion matric image, and record it to summary file
        # self.summary_record.add_value('image', 'image0', cb_params.train_dataset_element[0])
        self.summary_record.add_value('scalar', 'loss', cb_params.net_outputs)
        self.summary_record.record(cb_params.cur_step_num)


class StateCallback(Callback):
    def __init__(self, trainer_state, optimizer, train_steps, report_steps=1, summary_dir=None):
        self.trainer_state = trainer_state
        self.optimizer = optimizer
        self.report_steps = report_steps
        self.summary_dir = summary_dir
        self.summary_record = None
        self.train_steps = train_steps
        self.cur_step_num = 0

    def step_end(self, run_context):
        cb_params = run_context.original_args()
        if not hasattr(cb_params, 'cur_epoch_num'):
            cb_params.cur_epoch_num = 0
        self.cur_step_num = cb_params.cur_epoch_num
        loss = cb_params.net_outputs

        if isinstance(loss, (tuple, list)):
            if isinstance(loss[0], Tensor) and isinstance(loss[0].asnumpy(), np.ndarray):
                loss = loss[0]

        if isinstance(loss, Tensor) and isinstance(loss.asnumpy(), np.ndarray):
            loss = float(np.mean(loss.asnumpy()))

        cur_step_in_epoch = (cb_params.cur_step_num - 1) % cb_params.batch_num + 1

        if isinstance(loss, float) and (np.isnan(loss) or np.isinf(loss)):
            raise ValueError("epoch: {} step: {}. Invalid loss, terminating training.".format(
                cb_params.cur_epoch_num, cur_step_in_epoch))

        #In disaster recovery scenario, the cb_params.cur_step_num may be rollback to previous step
        # and be less than self._last_print_time, so self._last_print_time need to be updated.
        # if self._per_print_times != 0 and (cb_params.cur_step_num <= self._last_print_time):
        #     while cb_params.cur_step_num <= self._last_print_time:
        #         self._last_print_time -=\
        #             max(self._per_print_times, cb_params.batch_num if cb_params.dataset_sink_mode else 1)

        # if self._per_print_times != 0 and (cb_params.cur_step_num - self._last_print_time) >= self._per_print_times:
        #     self._last_print_time = cb_params.cur_step_num
            # print("epoch: %s step: %s, loss is %s" % (cb_params.cur_epoch_num, cur_step_in_epoch, loss), flush=True)

        lrs = get_lr(self.optimizer)
        if not isinstance(lrs, tuple):
            lrs = (lrs, )
        lr = sum(lrs).asnumpy() / len(lrs)
        if hasattr(cb_params.network, 'loss_scale'):
            loss_scale = getattr(cb_params.network, 'loss_scale', 1) * 1
            loss_scale = loss_scale.asnumpy()
        else:
            loss_scale = 1
        if hasattr(cb_params.train_dataset, 'batch_size'):
            batch_size = getattr(cb_params.train_dataset, 'batch_size', 1)
        else:
            batch_size = 1
        self.trainer_state.update_step(loss, batch_size, loss_scale, lr=lr)
        cb_params = run_context.original_args()

        if cb_params.cur_step_num % self.report_steps == 0:
            self.trainer_state.report_state()
            if self.summary_record is not None:
                self.summary_record.add_value('scalar', 'loss', ms.Tensor(loss))
                self.summary_record.add_value('scalar', 'learning rate', ms.Tensor(lr))
                self.summary_record.add_value('scalar', 'loss_scale', ms.Tensor(loss_scale))
                self.summary_record.record(cb_params.cur_step_num)
        if cb_params.cur_step_num >= self.train_steps:
            run_context.request_stop()

    def __enter__(self):
        # init you summary record in here, when the train script run, it will be inited before training
        if self.summary_dir:
            self.summary_record = ms.train.summary.SummaryRecord(self.summary_dir)
        return self

    def __exit__(self, *exc_args):
        # Note: you must close the summary record, it will release the process pool resource
        # else your training script will not exit from training.
        if self.summary_record:
            self.summary_record.close()


class ModelCheckpointWithBest(ModelCheckpoint):
    def __init__(self, trainer_state, prefix='CKP', directory=None, config=None, load_checkpoint_path=None):
        super().__init__(prefix=prefix, directory=directory, config=config)
        self.trainer_state = trainer_state
        self.best_metric = self.trainer_state.best_metric
        self.load_checkpoint_path = load_checkpoint_path

    def save_best(self, cb_params):
        if self.trainer_state.best_metric > self.best_metric:
            self.best_metric = self.trainer_state.best_metric
            ckpt_path = os.path.join(self._directory, 'best.ckpt')
            if os.path.exists(ckpt_path):
                os.remove(ckpt_path)
            self._save(cb_params, ckpt_path)

    def _save_ckpt(self, cb_params, force_to_save=False):
        """Save checkpoint files."""
        if cb_params.cur_step_num == self._last_triggered_step:
            return

        # if param is cache enable, flush data from cache to host before save_ckpt
        if self._need_flush_from_cache:
            self._flush_from_cache(cb_params)

        save_ckpt = self._check_save_ckpt(cb_params, force_to_save)
        step_num_in_epoch = int((cb_params.cur_step_num - 1) % cb_params.batch_num + 1)

        if save_ckpt:
            cur_ckpoint_file = self._prefix + "-" + str(cb_params.cur_step_num) + "_" \
                + str(cb_params.cur_epoch_num) + ".ckpt"
            # update checkpoint file list.
            self._manager.update_ckpoint_filelist(self._directory, self._prefix)
            # keep checkpoint files number equal max number.
            if self._config.keep_checkpoint_max and 0 < self._config.keep_checkpoint_max <= self._manager.ckpoint_num:
                self._manager.remove_oldest_ckpoint_file()
            elif self._config.keep_checkpoint_per_n_minutes and self._config.keep_checkpoint_per_n_minutes > 0:
                self._cur_time_for_keep = time.time()
                if (self._cur_time_for_keep - self._last_time_for_keep) \
                        < self._config.keep_checkpoint_per_n_minutes * 60:
                    self._manager.keep_one_ckpoint_per_minutes(self._config.keep_checkpoint_per_n_minutes,
                                                               self._cur_time_for_keep)

            # generate the new checkpoint file and rename it.
            # global _save_dir
            # _save_dir = self._directory
            self._save(cb_params, cur_ckpoint_file)
            self.save_best(cb_params)

    def _save(self, cb_params, cur_ckpoint_file):
        cur_file = os.path.join(self._directory, cur_ckpoint_file)
        self._last_time_for_keep = time.time()
        self._last_triggered_step = cb_params.cur_step_num

        if context.get_context("enable_ge"):
            _set_cur_net(cb_params.train_network)
            cb_params.train_network.exec_checkpoint_graph()
        if "epoch_num" in self._append_dict:
            self._append_dict["epoch_num"] = self._append_epoch_num + cb_params.cur_epoch_num
        if "step_num" in self._append_dict:
            self._append_dict["step_num"] = self._append_step_num + cb_params.cur_step_num
        network = self._config.saved_network if self._config.saved_network is not None else cb_params.train_network
        save_checkpoint(network, cur_file, self._config.integrated_save, self._config.async_save,
                        self._append_dict, self._config.enc_key, self._config.enc_mode)

        self._latest_ckpt_file_name = cur_file


class EvalResultsCallback(Callback):
    """
        Get results of evaluation steps
        the result is a Tuple of (Loss, net_outputs, label)
    """
    def __init__(self, result_fn='argmax') -> None:
        super().__init__()
        self.results = {}
        result_fns = {
            'argmax' : ms.ops.Argmax()
        }
        self.result_fn = result_fns.get(result_fn, None)
        parallel_mode = context.get_auto_parallel_context("parallel_mode")
        self.reduce_flag = False
        if parallel_mode in [ParallelMode.DATA_PARALLEL, ParallelMode.HYBRID_PARALLEL]:
            self.reduce_flag = True
            # self.allreduce = P.AllReduce()
            self.allgather = AllGather()
        else:
            self.allgather = None

    def step_end(self, run_context):
        loss, net_output, label = run_context.original_args().net_outputs
        # if self.allgather is not None:
        #     label = self.allgather(label)
        if self.result_fn is not None:
            net_output = self.result_fn(net_output)
        loss = loss.view((1, -1))
        loss = loss.asnumpy()
        net_output = net_output.asnumpy()
        label = label.asnumpy()
        if len(self.results) == 0:
            self.results['loss'] = loss
            self.results['net_output'] = net_output
            self.results['label'] = label
        else:
            self.results['loss'] = np.concatenate([self.results['loss'], loss], axis=0)
            self.results['net_output'] = np.concatenate([self.results['net_output'], net_output], axis=0)
            self.results['label'] = np.concatenate([self.results['label'], label], axis=0)


def get_lr(optimizer):
    """
    The optimizer calls this interface to get the learning rate for the current step. User-defined optimizers based
    on :class:`mindspore.nn.Optimizer` can also call this interface before updating the parameters.

    Returns:
        float, the learning rate of current step.
    """
    lr = optimizer.learning_rate
    if optimizer.dynamic_lr:
        if optimizer.is_group_lr:
            lr = ()
            for learning_rate in optimizer.learning_rate:
                current_dynamic_lr = learning_rate(optimizer.global_step)
                lr += (current_dynamic_lr,)
        else:
            lr = optimizer.learning_rate(optimizer.global_step)

    return lr


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================


import math
import torch
from torch.optim import Optimizer
from torch.nn.utils import clip_grad_norm_

def warmup_cosine(step, total, warmup=0.002, ends = 0):
    x = step/total
    x = x-int(x)
    if x < warmup:
        return x/warmup
    return 0.5 * (1.0 + math.cos(math.pi * x))

def warmup_constant(step, total, warmup=0.002, ends = 0):
    x = step/total
    x = x-int(x)
    if x < warmup:
        return x/warmup
    return 1.0

def warmup_linear(step, total, warmup=0.002, ends = 0):
    x = step/total
    x = x-int(x)
    if x < warmup:
        return x/warmup
    return (1-ends)*(1.0 - x) + ends

def warmup_linear_cosine(step, total, warmup=0.002, ends = 0):
    x = step/total
    x = x-int(x)
    if x < warmup:
        return x/warmup
    return (1-ends)*max(0.5*(1+math.cos(math.pi*(x-warmup)/(1-warmup))), 0) + ends

def warmup_cyclic_linear_cosine(step, total, warmup=0.002, ends = 0):
    x = step/total
    if x < warmup:
        return x/warmup
    total = total - int(total*warmup)
    step = step - int(total*warmup)
    n_epoch = 4
    period = total//n_epoch
    k = step//period
    s = 1-k/n_epoch + 1/(2*n_epoch)*(math.pow(-1, k)*math.cos(math.pi*step/period)-1)
    return (1-ends)*max(s, 0) + ends

def warmup_linear_shift(step, total, warmup=0.002, ends = 0):
    x = step/total
    x = x-int(x)
    if x < warmup:
        return x/warmup
    return (1-ends)*(1.0 - (x-warmup)/(1-warmup)) + ends

SCHEDULES = {
    'warmup_cosine':warmup_cosine,
    'warmup_constant':warmup_constant,
    'warmup_linear':warmup_linear,
    'warmup_linear_cosine':warmup_linear_cosine,
    'warmup_cyclic_linear_cosine':warmup_cyclic_linear_cosine,
    'warmup_linear_shift':warmup_linear_shift,
}



"""
Official evaluation script for ReCoRD v1.0.
(Some functions are adopted from the SQuAD evaluation script.)
"""

from __future__ import print_function
from collections import Counter
import string
import re
from loguru import logger

__all__=['normalize_answer', 'evaluate']


def normalize_answer(s):
    """Lower text and remove punctuation, articles and extra whitespace."""
    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(prediction, ground_truth):
    prediction_tokens = normalize_answer(prediction).split()
    ground_truth_tokens = normalize_answer(ground_truth).split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def exact_match_score(prediction, ground_truth):
    return normalize_answer(prediction) == normalize_answer(ground_truth)


def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
    scores_for_ground_truths = []
    for ground_truth in ground_truths:
        score = metric_fn(prediction, ground_truth)
        scores_for_ground_truths.append(score)
    return max(scores_for_ground_truths)


def evaluate(answers, predictions):
    f1 = exact_match = total = 0
    correct_ids = []
    for qid in answers:
        total += 1
        if qid not in predictions:
            message = 'Unanswered question {} will receive score 0.'.format(qid)
            logger.warning(message)
            continue

        ground_truths = list(map(lambda x: x['text'], answers[qid]))
        prediction = predictions[qid]

        _exact_match = metric_max_over_ground_truths(exact_match_score, prediction, ground_truths)
        if int(_exact_match) == 1:
            correct_ids.append(qid)
        exact_match += _exact_match

        f1 += metric_max_over_ground_truths(f1_score, prediction, ground_truths)

    exact_match = 100.0 * exact_match / total
    f1 = 100.0 * f1 / total

    return {'exact_match': exact_match, 'f1': f1}



"""
Byte pair encoding utilities from GPT-2.

Original source: https://github.com/openai/gpt-2/blob/master/src/encoder.py
Original license: MIT
"""

from functools import lru_cache
import json
import random
import unicodedata

try:
    import regex as re
except ImportError:
    raise ImportError('Please install regex with: pip install regex')

@lru_cache()
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = list(range(ord("!"), ord("~")+1))+list(range(ord("¡"), ord("¬")+1))+list(range(ord("®"), ord("ÿ")+1))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8+n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))

def get_pairs(word):
    """Return set of symbol pairs in a word.
    Word is represented as tuple of symbols (symbols being variable-length strings).
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs

class Encoder:

    def __init__(self, encoder, bpe_merges, errors='replace'):
        self.encoder = encoder
        self.decoder = {v:k for k,v in self.encoder.items()}
        self.errors = errors # how to handle errors in decoding
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v:k for k, v in self.byte_encoder.items()}
        self.bpe_ranks = dict(zip([tuple(k) for k in bpe_merges], range(len(bpe_merges))))
        self.cache = {}
        self.random = random.Random(0)

        # Should haved added re.IGNORECASE so BPE merges can happen for capitalized versions of contractions
        self.pat = re.compile(r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+""")

    def bpe(self, token):
        if token in self.cache:
            return self.cache[token]
        word = tuple(token)
        pairs = get_pairs(word)

        if not pairs:
            return token

        while True:
            bigram = min(pairs, key = lambda pair: self.bpe_ranks.get(pair, float('inf')))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word)-1 and word[i+1] == second:
                    new_word.append(first+second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = ' '.join(word)
        self.cache[token] = word
        return word
    
    def split_to_words(self, text):
      return list(re.findall(self.pat, text))

    def encode(self, text):
        bpe_tokens = []
        for token in self.split_to_words(text):
            token = ''.join(self.byte_encoder[b] for b in token.encode('utf-8'))
            bpe_tokens.extend(self.encoder[bpe_token] for bpe_token in self.bpe(token).split(' '))
        return bpe_tokens

    def decode(self, tokens):
        text = ''.join([self.decoder[token] for token in tokens])
        text = bytearray([self.byte_decoder[c] for c in text]).decode('utf-8', errors=self.errors)
        return text

def get_encoder(encoder, vocab):
    return Encoder(
        encoder=encoder,
        bpe_merges=vocab,
    )

def _is_whitespace(char):
    """Checks whether `chars` is a whitespace character."""
    # \t, \n, and \r are technically contorl characters but we treat them
    # as whitespace since they are generally considered as such.
    if char == " " or char == "\t" or char == "\n" or char == "\r":
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False

def _is_control(char):
    """Checks whether `chars` is a control character."""
    # These are technically control characters but we count them as whitespace
    # characters.
    if char == "\t" or char == "\n" or char == "\r":
        return False
    cat = unicodedata.category(char)
    if cat.startswith("C"):
        return True
    return False

def _is_punctuation(char):
    """Checks whether `chars` is a punctuation character."""
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
            (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================


import argparse
from ..utils import boolean_string

__all__ = ['get_args']

def get_args():
  parser=argparse.ArgumentParser(add_help=False, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  group = parser.add_argument_group(title='Trainer', description='Parameters for the distributed trainer')
  group.add_argument('--accumulative_update',
            type=int,
            default=1,
            help="Number of updates steps to accumulate before performing a backward/update pass.")

  group.add_argument('--loss_scale_value',
            type=int,
            default=1024,
            help="initial loss scale value")

  group.add_argument('--scale_factor',
            type=int,
            default=4,
            help="loss scale factor")

  group.add_argument('--scale_window',
            type=int,
            default=1000,
            help="loss window")

  group.add_argument('--enable_global_norm',
            type=bool,
            default=False,
            help="enable global norm")
  
  group.add_argument("--report_interval",
            default=1,
            type=int,
            help="Interval steps for state report.")

  group.add_argument("--local_rank",
            type=int,
            default=0,
            help="local_rank for distributed training on gpus")

  group.add_argument('--data_workers',
            type=int,
            default=4,
            help="The workers to load data.")

  group.add_argument('--seed',
            type=int,
            default=1234,
            help="random seed for initialization")

  group.add_argument("--train_batch_size",
            default=64,
            type=int,
            help="Total batch size for training.")

  group.add_argument("--world_size",
            type=int,
            default=-1,
            help="[Internal] The world size of distributed training. Internal usage only!! To the world size of the program, you need to use environment. 'WORLD_SIZE'")

  group.add_argument("--rank",
            type=int,
            default=-1,
            help="[Internal] The rank id of current process. Internal usage only!! To the rank of the program, you need to use environment. 'RANK'")

  group.add_argument("--master_ip",
            type=str,
            default=None,
            help="[Internal] The ip address of master node. Internal usage only!! To the master IP of the program, you need to use environment. 'MASTER_ADDR'")

  group.add_argument("--master_port",
            type=str,
            default=None,
            help="[Internal] The port of master node. Internal usage only!! To the master IP of the program, you need to use environment. 'MASTER_PORT'")

  return parser



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import math
import numpy as np
from bisect import bisect


class NGramMaskGenerator:
  """
  Mask ngram tokens
  https://github.com/zihangdai/xlnet/blob/0b642d14dd8aec7f1e1ecbf7d6942d5faa6be1f0/data_utils.py
  """
  def __init__(self, tokenizer, mask_lm_prob=0.15, max_seq_len=512, max_preds_per_seq=None, max_gram = 1, keep_prob = 0.1, mask_prob=0.8, **kwargs):
    self.tokenizer = tokenizer
    self.mask_lm_prob = mask_lm_prob
    self.keep_prob = keep_prob
    self.mask_prob = mask_prob
    assert self.mask_prob+self.keep_prob<=1, f'The prob of using [MASK]({mask_prob}) and the prob of using original token({keep_prob}) should between [0,1]'
    self.max_preds_per_seq = max_preds_per_seq
    if max_preds_per_seq is None:
      self.max_preds_per_seq = math.ceil(max_seq_len * mask_lm_prob / 10) * 10

    self.max_gram = max(max_gram, 1)
    self.mask_window = int(1/mask_lm_prob) # make ngrams per window sized context
    self.vocab_words = list(tokenizer._vocab.keys())

  def mask_tokens(self, tokens, rng, **kwargs):
    special_tokens = self.tokenizer.all_special_tokens# ['[MASK]', '[CLS]', '[SEP]', '[PAD]', '[UNK]'] # + self.tokenizer.tokenize(' ')
    indices = [i for i in range(len(tokens)) if tokens[i] not in special_tokens]
    ngrams = np.arange(1, self.max_gram + 1, dtype=np.int64)
    pvals = 1. / np.arange(1, self.max_gram + 1)
    pvals /= pvals.sum(keepdims=True)

    unigrams = []
    for id in indices:
      if self.max_gram>1 and len(unigrams)>=1 and self.tokenizer.part_of_whole_word(tokens[id]):
        unigrams[-1].append(id)
      else:
        unigrams.append([id])
    
    num_to_predict = min(self.max_preds_per_seq, max(1, int(round(len(tokens) * self.mask_lm_prob))))
    mask_len = 0
    offset = 0
    mask_grams = np.array([False] * len(unigrams))
    while offset < len(unigrams):
      n = self._choice(rng, ngrams, p=pvals)
      ctx_size = min(n * self.mask_window, len(unigrams) - offset)
      m = rng.randint(0, ctx_size - 1)
      s = offset + m
      e = min(offset + m + n, len(unigrams))
      offset = max(offset + ctx_size, e)
      mask_grams[s: e] = True

    target_labels = [None] * len(tokens)
    w_cnt = 0
    for m, word in zip(mask_grams, unigrams):
      if m:
        for idx in word:
          label = self._mask_token(idx, tokens, rng, self.mask_prob, self.keep_prob)
          target_labels[idx] = label
          w_cnt += 1
        if w_cnt >= num_to_predict:
          break

    target_labels = [self.tokenizer.convert_tokens_to_ids(x) if x else 0 for x in target_labels]
    return tokens, target_labels

  def _choice(self, rng, data, p):
    cul = np.cumsum(p)
    x = rng.random()*cul[-1]
    id = bisect(cul, x)
    return data[id]

  def _mask_token(self, idx, tokens, rng, mask_prob, keep_prob):
    label = tokens[idx]
    mask = '[MASK]'
    rand = rng.random()
    if rand < mask_prob:
      new_label = mask
    elif rand < mask_prob+keep_prob:
      new_label = label
    else:
      new_label = rng.choice(self.vocab_words)

    tokens[idx] = new_label

    return label


# -*- coding: utf-8 -*-
from setuptools import setup
from setuptools import find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding='utf-8')

setup(name='MindsporeTrainer',
      version='0.1.4',
      description='Make Mindspore Training Easier',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Zhou Bo',
      author_email='baudzhou@outlook.com',
      url='https://github.com/baudzhou/MindsporeTrainer',
      license="MIT",
      classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Text Processing',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Processing',
      ],
      install_requires=[
                        'loguru',
                        'ujson',
                        'tqdm',
                        'scipy',
                        'scikit-learn',
                        ],
      keywords='Deep Learning, NLP, CV, Transformers, MindSpore',
      packages=find_packages(),
)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Note: Mixture of Expert (MoE) structure. This is an experimental interface that is subject to change or deletion.
"""
import math
import numpy as np
from mindspore.common.tensor import Tensor
import mindspore.common.dtype as mstype
import mindspore.communication.management as D
from mindspore._checkparam import Validator
from mindspore.ops import operations as P
from mindspore.ops import functional as F
from mindspore.ops.primitive import constexpr
from mindspore.nn.cell import Cell
from mindspore.nn.layer import Dense
from mindspore.context import ParallelMode
from mindspore.parallel._utils import _get_parallel_mode, _is_sharding_propagation
from .op_parallel_config import default_moeparallel_config

__all__ = [
    "MoEConfig"]


class MoEConfig:
    r"""
        The configuration of MoE (Mixture of Expert).

        Args:
            expert_num (int): The number of experts employed. Default: 1
            capacity_factor (float): The factor is used to indicate how much to expand expert capacity,
                which is >=1.0. Default: 1.1.
            aux_loss_factor (float): The factor is used to indicate how much the load balance loss (produced by the
                router) to be added to the entire model loss, which is < 1.0. Default: 0.05.
            num_experts_chosen (int): The number of experts is chosen by each token and it should not be larger
                than expert_num. Default: 1.
        Supported Platforms:
            ``Ascend``

        Examples:
            >>> from mindspore.nn.transformer import MoEConfig
            >>> moe_config = MoEConfig(expert_num=4, capacity_factor=5.0, aux_loss_factor=0.05, num_experts_chosen=1)
    """

    def __init__(self, expert_num=1, capacity_factor=1.1, aux_loss_factor=0.05,
                 num_experts_chosen=1):
        Validator.check_positive_int(expert_num, "expert_num")
        Validator.check_positive_float(capacity_factor, "capacity_factor")
        Validator.check_positive_float(aux_loss_factor, "aux_loss_factor")
        Validator.check_positive_int(num_experts_chosen, "num_experts_chosen")
        if capacity_factor < 1.0:
            raise ValueError(f"'capacity_factor' should be equal to or greater than 1.0, "
                             f"but got {capacity_factor}.")
        if aux_loss_factor >= 1.0:
            raise ValueError(f"'aux_loss_factor' should be less than 1.0, "
                             f"but got {aux_loss_factor}.")
        if num_experts_chosen > expert_num:
            raise ValueError(f"'num_experts_chosen' should not be larger than 'expert_num', "
                             f"but got {num_experts_chosen}.")
        self.expert_num = expert_num
        self.capacity_factor = capacity_factor
        self.aux_loss_factor = aux_loss_factor
        self.num_experts_chosen = num_experts_chosen


default_moe_config = MoEConfig()


def _check_moe_config(moe_config=None, parallel_config=None):
    """
        check if MoE with right configuration.
    """
    if not isinstance(moe_config, MoEConfig):
        raise TypeError(f"'moe_config' should be an instance of MoEConfig, but got {type(moe_config).__name__}.")
    use_moe = (moe_config.expert_num > 1)
    if use_moe is False:
        return
    if moe_config.expert_num % parallel_config.expert_parallel != 0:
        raise ValueError(f"When using MoE, the 'expert_num' in {type(moe_config).__name__} must be a multiple "
                         f"of 'expert_parallel' value in {type(parallel_config).__name__}, but got "
                         f"{moe_config.expert_num} for 'expert_num' and {parallel_config.expert_parallel} for "
                         f"'expert_parallel'.")

    device_num = D.get_group_size()
    if device_num % parallel_config.expert_parallel != 0:
        raise ValueError(f"device_num: {device_num} should be a multiple of expert_parallel: "
                         f"{parallel_config.expert_parallel}.")
    if parallel_config.data_parallel % parallel_config.expert_parallel != 0:
        raise ValueError(f"data parallel: {parallel_config.data_parallel} should be a multiple of "
                         f"expert_parallel: {parallel_config.expert_parallel} when using MoE.")
    if parallel_config.data_parallel * parallel_config.model_parallel > device_num:
        raise ValueError(f"The product of the data parallel: {parallel_config.data_parallel} and "
                         f"model parallel: {parallel_config.model_parallel} "
                         f"should be less than device_num: {device_num}.")


@constexpr
def calculate_expert_capacity(k, tokens_per_group, capacity_factor, expert_dim):
    return math.ceil(k * tokens_per_group * capacity_factor / expert_dim)


class MoE(Cell):
    """
    The mixture of experts (MoE) implementation. The implementation includes a router and a FeedForward layer.
    The router dispatches tokens to experts in FeedForward, then FeedForward does computation, and the final output is
    obtained by multiplying FeedForward's output and router's combine weight.

    Args:
        hidden_size (int): The dimension of the inputs.
        ffn_hidden_size (int): The intermediate hidden size.
        dropout_rate (float): The dropout rate for the second linear's output.
        hidden_act (str): The activation of the internal feedforward layer. Supports 'relu',
                         'relu6', 'tanh', 'gelu', 'fast_gelu', 'elu', 'sigmoid', 'prelu', 'leakyrelu', 'hswish',
                         'hsigmoid', 'logsigmoid' and so on. Default: gelu.
        param_init_type (dtype.Number): The parameter initialization type. Can be dtype.float32 or dtype.float16.
        moe_config(MoEConfig): The configuration of MoE (Mixture of Expert). Default is an instance of MoEConfig with
            default values. Please see `MoEConfig`.
        parallel_config(MoEParallelConfig): The parallel config for MoE, see `MoEParallelConfig`.
            Default `default_moeparallel_config`, an instance of `MoEParallelConfig` with default args.

    Inputs:
        - **x** (Tensor) - should be `[batch, seq_length, hidden_size]`. Float tensor.

    Outputs:
        Tensor, the output of this layer after mapping. The shape is `[batch, seq_length, hidden_size]`.
    """

    def __init__(self, hidden_size,
                 ffn_hidden_size,
                 dropout_rate,
                 hidden_act='gelu',
                 param_init_type=mstype.float32,
                 moe_config=default_moe_config,
                 parallel_config=default_moeparallel_config):
        super(MoE, self).__init__()
        if _get_parallel_mode() in (ParallelMode.AUTO_PARALLEL,) and _is_sharding_propagation():
            self.hidden_size = hidden_size
            self.expert_dim = moe_config.expert_num
            self.capacity_factor = moe_config.capacity_factor
            self.aux_loss_factor = moe_config.aux_loss_factor
            self.num_experts_chosen = moe_config.num_experts_chosen
            self.dp_group = parallel_config.data_parallel
            self.dp = parallel_config.data_parallel
            self.ep = parallel_config.expert_parallel
            from .transformer import FeedForward

            self.ffn = FeedForward(hidden_size=hidden_size,
                                   ffn_hidden_size=ffn_hidden_size,
                                   dropout_rate=dropout_rate,
                                   hidden_act=hidden_act,
                                   expert_num=self.expert_dim,
                                   param_init_type=param_init_type,
                                   parallel_config=parallel_config)
            self.reshape = P.Reshape()
            self.shape = P.Shape()
            self.transpose_2dim = P.Transpose().shard(((self.dp, 1),))
            self.transpose_2dim_ep = P.Transpose().shard(((self.ep, 1),))
            self.transpose_3dim = P.Transpose().shard(((self.dp, 1, 1),))
            self.transpose_4dim_ep = P.Transpose().shard(((self.ep, 1, 1, 1),))
            self.batch_mm = P.BatchMatMul().shard(((self.dp, 1, 1), (self.dp, 1, 1)))
            self.batch_mm2 = P.BatchMatMul().shard(((self.dp, 1, 1), (self.dp, 1, 1)))
            self.mul = P.Mul()
            self.router = Router(d_model=hidden_size, moe_config=moe_config, routing_policy=None,
                                 training=True, parallel_config=parallel_config)
            self.cast = P.Cast()
        else:
            self.hidden_size = hidden_size
            self.expert_dim = moe_config.expert_num
            self.capacity_factor = moe_config.capacity_factor
            self.aux_loss_factor = moe_config.aux_loss_factor
            self.num_experts_chosen = moe_config.num_experts_chosen
            self.dp_group = parallel_config.data_parallel
            self.dp = parallel_config.data_parallel
            self.ep = parallel_config.expert_parallel
            from .transformer import FeedForward

            self.ffn = FeedForward(hidden_size=hidden_size,
                                   ffn_hidden_size=ffn_hidden_size,
                                   dropout_rate=dropout_rate,
                                   hidden_act=hidden_act,
                                   expert_num=self.expert_dim,
                                   param_init_type=param_init_type,
                                   parallel_config=parallel_config)
            self.reshape = P.Reshape()
            self.shape = P.Shape()
            self.transpose_2dim = P.Transpose().shard(((self.dp, 1),))
            self.transpose_2dim_ep = P.Transpose().shard(((self.ep, 1),))
            self.transpose_3dim = P.Transpose().shard(((self.dp, 1, 1),))
            self.transpose_4dim_ep = P.Transpose().shard(((self.ep, 1, 1, 1),))
            self.batch_mm = P.BatchMatMul().shard(((self.dp, 1, 1), (self.dp, 1, 1)))
            self.batch_mm2 = P.BatchMatMul().shard(((self.dp, 1, 1), (self.dp, 1, 1)))
            self.mul = P.Mul().shard(((), ()))
            self.router = Router(d_model=hidden_size, moe_config=moe_config, routing_policy=None,
                                 training=True, parallel_config=parallel_config)
            self.cast = P.Cast()

    def construct(self, input_tensor):
        input_shape = F.shape(input_tensor)
        input_tensor = self.reshape(input_tensor, (-1, self.hidden_size))
        bs_and_dmodel = self.shape(input_tensor)
        tokens_per_group = bs_and_dmodel[0] // self.dp_group
        input_tensor = self.reshape(input_tensor, (self.dp_group, tokens_per_group, self.hidden_size))

        expert_capacity = calculate_expert_capacity(self.num_experts_chosen, tokens_per_group,
                                                    self.capacity_factor, self.expert_dim)
        # dispatch_tensor's shape: (self.dp_group, tokens_per_group, self.expert_dim, expert_capacity)
        # combine_tensor's shape: (self.dp_group, tokens_per_group, self.expert_dim, expert_capacity)
        dispatch_tensor, combine_tensor, aux_loss = self.router(input_tensor)

        # after transpose, input_tensor's shape: (self.dp_group, self.hidden_size, tokens_per_group)
        input_tensor = self.transpose_3dim(input_tensor, (0, 2, 1))
        dispatch_tensor = self.reshape(dispatch_tensor, (self.dp_group, tokens_per_group,
                                                         self.expert_dim * expert_capacity))
        dispatch_tensor = self.cast(dispatch_tensor, F.dtype(input_tensor))
        # expert_input's shape: (self.dp_group, self.hidden_size, self.expert_dim * expert_capacity)
        expert_input = self.batch_mm(input_tensor, dispatch_tensor)
        expert_input = self.reshape(expert_input, (self.dp_group, self.hidden_size, self.expert_dim,
                                                   expert_capacity))
        # The following four ops are to implement transpose(expert_input, (2, 0, 3, 1)), for that a single transpose
        # has bad performance
        expert_input = self.reshape(expert_input, (self.dp_group * self.hidden_size,
                                                   self.expert_dim * expert_capacity))
        expert_input = self.transpose_2dim(expert_input, (1, 0))
        expert_input = self.reshape(expert_input, (self.expert_dim, expert_capacity, self.dp_group,
                                                   self.hidden_size))
        # expert_input's shape: (self.expert_dim, self.dp_group, expert_capacity, self.hidden_size)
        expert_input = self.transpose_4dim_ep(expert_input, (0, 2, 1, 3))
        expert_input = self.reshape(expert_input, (self.expert_dim * self.dp_group * expert_capacity,
                                                   self.hidden_size))

        # expert_output's shape: (self.expert_dim, self.dp_group*expert_capacity, self.hidden_size)
        expert_output = self.ffn(expert_input)
        expert_output = self.reshape(expert_output, (self.expert_dim, self.dp_group,
                                                     expert_capacity, self.hidden_size))
        # The following five ops are to implement transpose(expert_output, (1, 3, 0, 2)), for that a single transpose
        # has bad performance
        expert_output = self.reshape(expert_output, (self.expert_dim,
                                                     self.dp_group * expert_capacity * self.hidden_size))
        expert_output = self.transpose_2dim_ep(expert_output, (1, 0))
        expert_output = self.reshape(expert_output, (self.dp_group, expert_capacity,
                                                     self.hidden_size * self.expert_dim))
        expert_output = self.transpose_3dim(expert_output, (0, 2, 1))
        # expert_output's shape: (self.dp_group, self.hidden_size, self.expert_dim, expert_capacity)
        expert_output = self.reshape(expert_output, (self.dp_group, self.hidden_size, self.expert_dim,
                                                     expert_capacity))
        expert_output = self.reshape(expert_output, (self.dp_group, self.hidden_size,
                                                     self.expert_dim * expert_capacity))
        combine_tensor = self.reshape(combine_tensor, (self.dp_group, tokens_per_group,
                                                       self.expert_dim * expert_capacity))
        # combine_tensor's shape: (self.dp_group, self.expert_dim*expert_capacity, tokens_per_group)
        combine_tensor = self.transpose_3dim(combine_tensor, (0, 2, 1))
        combine_tensor = self.cast(combine_tensor, F.dtype(expert_output))

        # combined_output's shape: (self.dp_group, self.hidden_size, tokens_per_group)
        combined_output = self.batch_mm2(expert_output, combine_tensor)
        # combined_output's shape: (self.dp_group, tokens_per_group, self.hidden_size)
        combined_output = self.transpose_3dim(combined_output, (0, 2, 1))
        combined_output = self.reshape(combined_output, (bs_and_dmodel[0], bs_and_dmodel[1]))
        combined_output = self.reshape(combined_output, input_shape)

        aux_loss = self.mul(self.aux_loss_factor, aux_loss)
        return combined_output, aux_loss


class Router(Cell):
    r"""
        A router backbone used to calculate logits of each token, which should be cascaded by router implementations
        mapping tokens to experts.
        when moe_config.num_experts_chosen = 1, use top1 routing;
        when moe_config.num_experts_chosen > 1, use topk routing

        Args:
            d_model (int): The hidden size of each token.
            moe_config(MoEConfig): The configuration of MoE (Mixture of Expert).
            routing_policy: The policy of mapping tokens to experts. Default: topkRouter
            training (bool): The value indicating whether is in training phase.
            parallel_config: The parallel-related configuration.
        Inputs:
            - **input_tensor** (Tensor) - Tensor of shape :math:`(expert\_parallel, tokens\_per\_device,
            hidden\_size)`.

        Outputs:
            Tensor of shape :math:`(expert\_parallel, tokens\_per\_device, expert\_dim)`.
    """

    def __init__(self,
                 d_model,
                 moe_config,
                 routing_policy=None,
                 training=True,
                 parallel_config=None):
        super(Router, self).__init__()
        if _get_parallel_mode() in (ParallelMode.AUTO_PARALLEL,) and _is_sharding_propagation():
            self.d_model = d_model
            self.expert_dim = moe_config.expert_num
            self.capacity_factor = moe_config.capacity_factor
            self.num_experts_chosen = moe_config.num_experts_chosen
            self.training = training
            self.routing_policy = routing_policy
            self.noisy_policy = None  # candidate: ["jitter", "rsample", "None"]
            self.noisy_epsilon = 1e-2
            self.noise = Tensor(np.random.uniform(1 - self.noisy_epsilon, 1 + self.noisy_epsilon, (d_model,)))

            self.dense = Dense(in_channels=self.d_model, out_channels=self.expert_dim, has_bias=False)
            self.mul = P.Mul()
            self.cast = P.Cast()

            if self.routing_policy is None:
                self.router = TopkRouter(d_model=d_model, moe_config=moe_config, training=training,
                                         parallel_config=parallel_config)
            else:
                self.router = routing_policy
        else:
            dp = parallel_config.data_parallel
            self.d_model = d_model
            self.expert_dim = moe_config.expert_num
            self.capacity_factor = moe_config.capacity_factor
            self.num_experts_chosen = moe_config.num_experts_chosen
            self.training = training
            self.routing_policy = routing_policy
            self.noisy_policy = None  # candidate: ["jitter", "rsample", "None"]
            self.noisy_epsilon = 1e-2
            self.noise = Tensor(np.random.uniform(1 - self.noisy_epsilon, 1 + self.noisy_epsilon, (d_model,)))

            self.dense = Dense(in_channels=self.d_model, out_channels=self.expert_dim, has_bias=False)
            self.dense.matmul.shard(((dp, 1), (1, 1)))
            self.mul = P.Mul().shard(((dp, 1, 1), (dp,)))
            self.cast = P.Cast()

            if self.routing_policy is None:
                self.router = TopkRouter(d_model=d_model, moe_config=moe_config, training=training,
                                         parallel_config=parallel_config)
            else:
                self.router = routing_policy

    def construct(self, input_tensor):
        input_tensor = self.cast(input_tensor, mstype.float32)
        if self.noisy_policy == "jitter" and self.training:
            # Here, we temporarily implement the multiplicative jitter this way,
            # for the lack of UniforReal parallel operator.
            input_tensor = self.mul(input_tensor, self.noise)

        router_logits = self.dense(input_tensor)
        return self.router(router_logits)


class TopkRouter(Cell):
    r"""
        A router implementation which maps each tokens to the topk expert.

        Args:
            d_model (int): The hidden size of each token.
            moe_config(MoEConfig): The configuration of MoE (Mixture of Expert).
            training (bool): The value indicating whether is in training phase.
            config: The parallel-related configuration.
        Inputs:
            - **input_tensor** (Tensor) - Tensor of shape :math:`(expert\_parallel, tokens\_per\_device,
            hidden\_size)`.

        Outputs:
            Tensor of shape :math:`(expert\_parallel, tokens\_per\_device, expert\_dim, expert\_capacity)`,
            Tensor of shape :math:`(expert\_parallel, tokens\_per\_device, expert\_dim, expert\_capacity)`,
            Tensor of shape :math:`(1)`.
    """

    def __init__(self,
                 d_model,
                 moe_config,
                 training=True,
                 parallel_config=None):
        super(TopkRouter, self).__init__()
        if _get_parallel_mode() in (ParallelMode.AUTO_PARALLEL,) and _is_sharding_propagation():
            dp = parallel_config.data_parallel
            self.d_model = d_model
            self.expert_dim = moe_config.expert_num
            self.capacity_factor = moe_config.capacity_factor
            self.training = training
            self.dp_group = dp
            self.noisy_policy = None
            self.cast = P.Cast()
            self.reshape = P.Reshape()
            self.shape = P.Shape()
            self.softmax = P.Softmax(axis=-1)
            self.argmax = P.ArgMaxWithValue(axis=-1, keep_dims=False)
            self.num_experts_chosen = moe_config.num_experts_chosen
            self.onehot = P.OneHot()
            self.onehot2 = P.OneHot()
            self.onehot3 = P.OneHot()
            self.on_value = Tensor(1.0, mstype.float32)
            self.off_value = Tensor(0.0, mstype.float32)

            self.reduce_mean = P.ReduceMean(keep_dims=False)
            self.reduce_mean2 = P.ReduceMean(keep_dims=False)
            self.reduce_mean3 = P.ReduceMean(keep_dims=False)
            self.mul = P.Mul()
            self.mul2 = P.Mul()
            self.mul3 = P.Mul()
            self.mul4 = P.Mul()
            self.mul5 = P.Mul()
            self.mul6 = P.Mul()
            self.mul7 = P.Mul()
            self.mul8 = P.Mul().shard(((dp, 1, 1), (dp, 1, 1)))
            self.mul9 = P.Mul().shard(((dp, 1, 1, 1), (dp, 1, 1, 1)))
            self.not_equal = P.NotEqual()
            self.div1 = P.RealDiv()
            self.div2 = P.RealDiv()
            self.add = P.Add()
            self.add1 = P.Add()
            self.add2 = P.Add()
            self.add3 = P.Add()
            self.add4 = P.Add()
            self.sub = P.Sub()

            self.cumsum = P.CumSum(exclusive=True)
            self.less = P.Less()
            self.reduce_sum = P.ReduceSum(keep_dims=False)
            self.reduce_sum_keep = P.ReduceSum(keep_dims=True)
            self.reduce_sum_keep2 = P.ReduceSum(keep_dims=True)
            self.expand = P.ExpandDims()
            self.expand2 = P.ExpandDims()
        else:
            dp = parallel_config.data_parallel
            self.d_model = d_model
            self.expert_dim = moe_config.expert_num
            self.capacity_factor = moe_config.capacity_factor
            self.training = training
            self.dp_group = dp
            self.noisy_policy = None
            self.cast = P.Cast()
            self.reshape = P.Reshape()
            self.shape = P.Shape()
            self.softmax = P.Softmax(axis=-1).shard(((dp, 1, 1,),))
            self.argmax = P.ArgMaxWithValue(axis=-1, keep_dims=False).shard(((dp, 1, 1),))
            self.num_experts_chosen = moe_config.num_experts_chosen
            self.onehot = P.OneHot().shard(((dp, 1, 1), (), ()))
            self.onehot2 = P.OneHot().shard(((dp, 1, 1), (), ()))
            self.onehot3 = P.OneHot().shard(((dp, 1, 1, 1), (), ()))
            self.on_value = Tensor(1.0, mstype.float32)
            self.off_value = Tensor(0.0, mstype.float32)

            self.reduce_mean = P.ReduceMean(keep_dims=False).shard(((dp, 1, 1),))
            self.reduce_mean2 = P.ReduceMean(keep_dims=False).shard(((dp, 1, 1),))
            self.reduce_mean3 = P.ReduceMean(keep_dims=False).shard(((dp, 1),))
            self.mul = P.Mul().shard(((dp, 1), (dp, 1)))
            self.mul2 = P.Mul().shard(((), ()))
            self.mul3 = P.Mul().shard(((), ()))
            self.mul4 = P.Mul().shard(((dp, 1, 1), (dp, 1, 1)))
            self.mul5 = P.Mul().shard(((dp, 1, 1), (dp, 1, 1)))
            self.mul6 = P.Mul().shard(((dp, 1), (dp, 1)))
            self.mul7 = P.Mul().shard(((dp, 1), (dp, 1)))
            self.mul8 = P.Mul().shard(((dp, 1, 1), (dp, 1, 1)))
            self.mul9 = P.Mul().shard(((dp, 1, 1, 1), (dp, 1, 1, 1)))
            self.not_equal = P.NotEqual().shard(((dp, 1, 1, 1), ()))
            self.div1 = P.RealDiv().shard(((dp, 1, 1), (dp, 1, 1)))
            self.div2 = P.RealDiv().shard(((dp, 1, 1, 1), (dp, 1, 1, 1)))
            self.add = P.Add().shard(((dp, 1, 1), (dp, 1, 1)))
            self.add1 = P.Add().shard(((dp, 1, 1), ()))
            self.add2 = P.Add().shard(((dp, 1, 1, 1), (dp, 1, 1, 1)))
            self.add3 = P.Add().shard(((dp, 1), (dp, 1)))
            self.add4 = P.Add().shard(((dp, 1, 1, 1), ()))
            self.sub = P.Sub().shard(((), (dp, 1, 1)))

            self.cumsum = P.CumSum(exclusive=True).shard(((dp, 1, 1),))
            self.less = P.Less().shard(((dp, 1, 1), ()))
            self.reduce_sum = P.ReduceSum(keep_dims=False).shard(((dp, 1, 1),))
            self.reduce_sum_keep = P.ReduceSum(keep_dims=True).shard(((dp, 1, 1),))
            self.reduce_sum_keep2 = P.ReduceSum(keep_dims=True).shard(((dp, 1, 1, 1),))
            self.expand = P.ExpandDims().shard(((dp, 1),))
            self.expand2 = P.ExpandDims().shard(((dp, 1, 1),))

    def _auxiliary_loss(self, expert_mask, router_prob):
        """
        Computing the load balance loss.
        """
        # density_1's shape: (dp_group, self.expert_dim)
        density_1 = self.reduce_mean(expert_mask, 1)
        # density_1_proxy's shape: (dp_group, self.expert_dim)
        density_1_proxy = self.reduce_mean2(router_prob, 1)
        loss = self.mul(density_1, density_1_proxy)
        loss = self.reduce_mean3(loss)
        loss = self.mul3(self.mul2(loss, self.expert_dim), self.expert_dim)
        return loss

    def _maskout_overflowed_tokens(self, expert_mask, expert_capacity, expert_gate, last_num, expert_chosen_index):
        """
        Keeping only the tokens that fit within expert_capacity.
        """
        cumsum = self.cumsum(expert_mask, 1)
        if expert_chosen_index > 0:
            cumsum = self.add(cumsum, last_num)
        # position_in_expert's shape: (dp_group, tokens_per_group, self.expert_dim)
        position_in_expert = self.mul4(cumsum, expert_mask)
        less_result = self.less(position_in_expert, expert_capacity)
        # expert_mask's shape: (dp_group, tokens_per_group, self.expert_dim)
        expert_mask = self.mul5(less_result, expert_mask)
        # expert_mask_flat's shape: (dp_group, tokens_per_group)
        expert_mask_flat = self.reduce_sum(expert_mask, -1)

        # Mask out the experts that have overflowed the expert_capacity.
        # expert_gate's shape: (dp_group, tokens_per_group)
        expert_gate = self.mul6(expert_gate, expert_mask_flat)
        output = (expert_mask, expert_gate, expert_mask_flat, position_in_expert)
        return output

    def construct(self, router_logits):
        router_logits_shape = self.shape(router_logits)
        router_logits = self.reshape(router_logits, (-1, router_logits_shape[-1]))
        logits_shape = self.shape(router_logits)
        tokens_per_group = logits_shape[0] // self.dp_group
        expert_capacity = calculate_expert_capacity(self.num_experts_chosen, tokens_per_group, self.capacity_factor,
                                                    self.expert_dim)
        router_logits = self.reshape(router_logits, (self.dp_group, tokens_per_group, self.expert_dim))

        accum_expert_mask = 0
        accum_expert_gate = 0
        loss = 0
        mask_count = 0
        accum_combine_tensor = 0
        # Probabilities for each token of what expert is should be sent to
        router_prob = self.softmax(router_logits)

        for expert_chosen_index in range(self.num_experts_chosen):
            # for each token, set the router_prob of the selected experts to zero
            router_prob = self.mul4(router_prob, self.sub(self.on_value, accum_expert_mask))
            # shape is : (dp_group, tokens_per_group)
            expert_index, expert_gate = self.argmax(router_prob)
            # expert_mask's shape: (dp_group, tokens_per_group, self.expert_dim)
            expert_mask = self.onehot(expert_index, self.expert_dim, self.on_value, self.off_value)
            # renormalize the rest prob to be of sum 1
            router_prob_normal = self.div1(router_prob, self.add1(self.reduce_sum_keep(router_prob, -1), 1e-9))

            # the balance loss is computed at each routing step
            loss += self._auxiliary_loss(expert_mask, router_prob_normal)

            output = self._maskout_overflowed_tokens(expert_mask, expert_capacity, expert_gate,
                                                     mask_count, expert_chosen_index)
            expert_mask, expert_gate, expert_mask_flat, position_in_expert = output[0], output[1], output[2], output[3]
            accum_expert_mask = self.add(accum_expert_mask, expert_mask)
            accum_expert_gate = self.add3(accum_expert_gate, expert_gate)
            mask_count = self.add(mask_count, self.reduce_sum_keep(expert_mask, 1))

            # combine_tensor's shape: (dp_group, tokens_per_group)
            combine_tensor = self.mul7(expert_gate, expert_mask_flat)
            # combine_tensor's shape: (dp_group, tokens_per_group, self.expert_dim)
            combine_tensor = self.mul8(self.expand(combine_tensor, -1),
                                       self.onehot2(expert_index, self.expert_dim, self.on_value, self.off_value))
            # combine_tensor's shape: (dp_group, tokens_per_group, self.expert_dim, self.expert_capacity)
            combine_tensor = self.mul9(self.expand2(combine_tensor, -1),
                                       self.onehot3(self.cast(position_in_expert, mstype.int32), expert_capacity,
                                                    self.on_value, self.off_value))
            accum_combine_tensor = self.add2(accum_combine_tensor, combine_tensor)

        # expert weights normalization
        combine_tensor_sum = self.reduce_sum_keep2(self.reduce_sum_keep2(accum_combine_tensor, -1), -2)
        accum_combine_tensor = self.div2(accum_combine_tensor, self.add4(combine_tensor_sum, 1e-9))
        # dispatch_tensor is of boolean type. Here, using NotEqual instead of Cast, for that 'Cast to bool' has
        # bad performance
        dispatch_tensor = self.not_equal(accum_combine_tensor, 0.0)
        return dispatch_tensor, accum_combine_tensor, loss



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
from .logger_util import *
from .argument_types import *
from .xtqdm import *

def get_dir():
    cwd = os.path.dirname(os.path.abspath(__file__))
    cwd = cwd.split(cwd, os.path.sep)